/* LeMoNe - Learning Module Network
 * 
 * Copyright (C) 2005-2008 VIB 
 * 
 * Written by Tom Michoel, Eric Bonnet, Anagha Joshi, Steven Maere
 *  
*/

package be.ugent.psb.ModuleNetwork;

import BiNGO.*;
import java.util.*;
import java.io.IOException;
import java.math.BigDecimal;
import cytoscape.data.annotation.OntologyTerm;


/**
 * Wrapper around the BiNGO library to generate GO statistics for each module of a 
 * ModuleNetwork object. 
 * 
 * @author stmae, erbon
 * 
 */
public class BiNGO {
	
	/**
	 * BingoParameters object.
	 */
    private BingoParameters bp;

    /**
     * BiNGO results file object.
     */
    private CreateBiNGOFile file;
    
    /**
     * ModuleNetwork object.
     */
    private ModuleNetwork M;
    
    /**
     * BiNGO object constructor
     * 
     * @param M: ModuleNetwork object
     * @param annotationFile annotation file name 
     * @param ontologyFile ontology file name
     * @param significance p-value threshold
     * @param annot_def ???
     * @param ontol_def ???
     * @return BiNGO object
     */
    public BiNGO(ModuleNetwork M, String annotationFile, String ontologyFile, String significance, boolean annot_def, boolean ontol_def) {

    	this.M = M;

    	// create bingoparameters object
    	try {
    		this.bp = new BingoParameters("");
    	} catch(IOException e){
    		System.out.println("Error: Can't create the BingoParameters object.");
    		System.exit(1);
    	}

    	// Override standards parameters with 
    	bp.setReferenceSet(BingoAlgorithm.GRAPH);
    	bp.setCategory(BingoAlgorithm.CATEGORY_CORRECTION);
    	bp.setOverOrUnder("Overrepresentation") ;
    	bp.setSignificance(new BigDecimal(significance)) ;                
    	
    	if(annot_def == true){
    		bp.setAnnotationFile(openResourceFile(annotationFile)) ;
    		bp.setSpecies(bp.getSpeciesNameFromFilename(annotationFile));
    	}
    	else{
    		bp.setAnnotationFile(annotationFile) ;
    		bp.setSpecies("custom");
    	}
    	
    	bp.setOntologyFile(openResourceFile(ontologyFile)) ;
    	bp.setAnnotation_default(annot_def) ;
    	bp.setOntology_default(ontol_def) ;
    	
    	AnnotationParser annParser = bp.initializeAnnotationParser();
    	
    	annParser.calculate() ;
    	
    	if (annParser.getStatus()) {
    		bp.setAnnotation(annParser.getAnnotation());
    		bp.setOntology(annParser.getOntology());
    		bp.setAlias(annParser.getAlias());
    	}
    	
    	if (bp.getAnnParser().getOrphans())
    		System.out.println("WARNING : Some category labels in the annotation file are not " +
    				"defined in the ontology. Please check the compatibility of these files. " +
    				"For now, these labels will be ignored and calculations will proceed.");
    }
    
    
    /**
     * Compute GO stats for every module of the ModuleNetwork object and write results to a file.
     * The module list used for the computations is the one pointed by the moduleSet property.
     * 
     * @param filename Output file name. Results are written in the current directory.
     * @param allNodes HashSet containing the reference list of genes. If empty, the complete list
     * of genes within the ModuleNetwork will be used as the reference.
     * @return nothing.
     */
    public void GOwrite(String filename, HashSet<String> allNodes) {

    	bp.setFileoutput(true);
    	//bp.setFileoutput_dir(bingoDir) ;

    	// if allNodes is empty, then fill it with all the ModuleNetwork genes
    	if (allNodes.size() == 0) {
    		System.out.println("Using module set genes as the reference set.");
    		for (Gene g : M.geneSet){
    			allNodes.add(g.name.toUpperCase());
    		}
    	}
    	bp.setAllNodes(allNodes);
    	bp.setCluster_name("");

    	// Create the bingo results file
    	file = new CreateBiNGOFile(
    			bp.getSignificance().toString(),
    			bp.getAnnotation(),
    			bp.getDeleteCodes(),
    			bp.getAlias(),
    			bp.getOntology(),
    			bp.getAnnotationFile().toString(),
    			bp.getOntologyFile().toString(),
    			bp.getTest() + "",
    			bp.getCorrection() + "",
    			bp.getOverOrUnder() + "",
    			//bp.getFileoutput_dir(),
    			System.getProperty("user"),
    			//bp.getCluster_name() + ".bgo",
    			filename,
    			bp.getReferenceSet() + "",
    			bp.getCategory() + "");

    	// loop over all modules 
    	for (int i=0;i<M.moduleSet.size();i++) {

    		String cluster_name = String.format("module_%02d", i);
    		
    		bp.setCluster_name(cluster_name);
    		HashSet<String> selectedNodes = new HashSet<String>();

    		// add module genes to selectedNodes 
    		for (Gene g : M.moduleSet.get(i).genes) {
    			selectedNodes.add(g.name);
    		}
    		bp.setSelectedNodes(selectedNodes);

    		// Compute GO statistics; results written to the bingo file on-the-fly
    		String name = Integer.toString(i);
    		
    		// function call for bingo 2007 version
    		//calculateGoEnrichment(bp.getSelectedNodes(), bp.getAllNodes(), name, i);
    		
    		// function call for 2008 version and above
    		calculateGoEnrichmentNew(bp.getSelectedNodes(), bp.getAllNodes(), name, i);
    	}
    }
    
    /**
     * Compute GO stats for all modules and set GO properties for genes and modules.
     *  
     * @param hashset containing the list of all genes used as a reference to compute GO statistics. If empty, the complete list
     * of genes within the ModuleNetwork will be used as the reference.
     * @return nothing.
     */
    public void GOmodules(HashSet<String> allNodes) {

    	// we don't want any output file
    	bp.setFileoutput(false);

    	// if allNodes is empty, then fill it with all the ModuleNetwork genes
    	if (allNodes.size() == 0) {
    		for (Gene g : M.geneSet){
    			allNodes.add(g.name.toUpperCase());
    		}
    	} 
    	bp.setAllNodes(allNodes);
    	bp.setCluster_name("");

    	// annotate all the genes
    	for (Gene g : M.geneSet) {

    		// get alias object
    		HashMap<String,HashSet<String>> alias = bp.getAlias();

    		// array for go labels.
    		//HashSet<String> cls = new HashSet<String>();
    		
    		// loop over all aliases for the gene
    		if (alias.get(g.name) != null) {
    			for (String id : alias.get(g.name)) {
    				// get all Go codes for this alias
    				for (int code : bp.getAnnotation().getClassifications(id) ) {
    					//cls.add(code.toString());
    					g.GOcat.add(bp.getOntology().getTerm(code));
    				}
    			}
    		}
    	}

    	// loop over all modules
    	for (int i=0;i<M.moduleSet.size();i++) {

    		String cluster_name = String.format("module_%02d", i);

    		bp.setCluster_name(cluster_name);
    		HashSet<String> selectedNodes = new HashSet<String>();

    		// add module genes to selectedNodes 
    		for (Gene g : M.moduleSet.get(i).genes) {
    			selectedNodes.add(g.name);
    		}
    		bp.setSelectedNodes(selectedNodes);

    		// Compute GO statistics and set GO properties for each module.

    		HashSet<String> noClassificationsSet = new HashSet<String>() ;
    		HashMap testMap = null;
    		HashMap correctionMap = null;
    		HashMap mapSmallX = null;
    		int[] classifications = getClassificationsFromVector(selectedNodes, noClassificationsSet) ;

    		BingoAlgorithm algorithm = new BingoAlgorithm(bp);
    		CalculateTestTask test = algorithm.calculate_distribution();
    		test.run();
    		testMap = test.getTestMap();
    		CalculateCorrectionTask correction = algorithm.calculate_corrections(testMap);
    		if (correction != null) {
    			correction.run() ;
    			correctionMap = correction.getCorrectionMap();
    		}
    		mapSmallX = test.getMapSmallX();

    		String[] ordenedKeySet = ordenKeys(correctionMap);

    		// add overrepresented GO cats to module;
    		// loop over categories ordered by p-values (lowest to highest)
    		for(String s: ordenedKeySet){
    			if(Double.parseDouble(correctionMap.get(s).toString()) < bp.getSignificance().doubleValue()){
    				OntologyTerm o = bp.getOntology().getTerm(Integer.parseInt(s));
    				
    				// add all GO terms below cutoff value to the module 
    				M.moduleSet.get(i).GOall.add(o);
    				
    				// only add those with lowest p-value for a certain branch of the tree
    				boolean ok = true;
    				// discard too small categories, usually set to 3
    				if (Integer.parseInt(mapSmallX.get(new Integer(s)).toString()) < 3){
    					ok = false;
    				}
    	            else { 
    	            	// don't add categories that belongs to the same branch
    	            	for(OntologyTerm p : M.moduleSet.get(i).GOsuper) {
    	            		if ( o.isAncestorOf(bp.getOntology(),p) || p.isAncestorOf(bp.getOntology(),o) ){
    	            			ok = false;
    	            		}
    	            	}
    				}
    				if (ok == true) {
    					M.moduleSet.get(i).GOsuper.add(o);    
    				}
    			}
    		}
    	}
    }

    
    private String openResourceFile(String name) {
        //System.out.println("return: "+bp.getClass().getResource("/" + name).toString());
    	return bp.getClass().getResource("/"+name).toString();
    }
    
    public int [] getClassificationsFromVector(HashSet canonicalNameVector, HashSet noClassificationsSet) {
        // HashSet for the classifications.
        HashSet classificationsVector = new HashSet();
        HashMap<String,HashSet<String>> alias = bp.getAlias();
        // array for go labels.
        int [] goLabelsName;
        
        // loop over all gene names
        Iterator it2 = canonicalNameVector.iterator();
        while(it2.hasNext()){     
            String name = it2.next() + "";
            // get all aliases for this gene name
            HashSet identifiers = alias.get(name) ;
            
            // array for go labels.
            HashSet<String> cls = new HashSet<String>();
            
            // loop over all aliases
            if(identifiers != null){
                Iterator it = identifiers.iterator() ;
                while(it.hasNext()){      
                    // get an [int] array of GO classifications for that gene
                	goLabelsName = bp.getAnnotation().getClassifications(it.next() + "");
                    for (int t = 0; t < goLabelsName.length; t++){
                        cls.add(goLabelsName[t] + "");
                    }
                }
            }
            
            // if the gene is not annotated, add it to unclassified
            if(cls.size() == 0){
                  noClassificationsSet.add(name);
            }
            Iterator it3 = cls.iterator();
            while(it3.hasNext()){
                  classificationsVector.add(it3.next() + "");
            }
            
        }      
        int [] classifications = new int [classificationsVector.size()];
        it2 = classificationsVector.iterator();
        int i = 0;
        while(it2.hasNext()){
            classifications[i] = Integer.parseInt(it2.next() + "");
            i++;
        }
        return classifications;
    }
    
//    public void calculateGoEnrichment(HashSet selectedNodes, HashSet allNodes, String name, int number) {
//        
//        HashMap testMap = null;
//        HashMap correctionMap = null;
//        HashMap mapSmallX = null;
//        HashMap mapSmallN = null;
//        int bigX = 0;
//        int bigN = 0;
//        HashSet noClassificationsSet = new HashSet() ;
//        int[] classifications = getClassificationsFromVector(selectedNodes, noClassificationsSet) ;
//       
//        BingoAlgorithm algorithm = new BingoAlgorithm(bp);
//        
//        CalculateTestTask test = algorithm.calculate_distribution();
//        test.run();
//		
//        testMap = test.getTestMap();
//        CalculateCorrectionTask correction = algorithm.calculate_corrections(testMap);
//        if (correction != null) {
//            correction.run() ;
//            correctionMap = correction.getCorrectionMap();
//        }
//        
//        mapSmallX = test.getMapSmallX();
//        mapSmallN = test.getMapSmallN();
//        bigX = test.getBigX();
//        bigN = test.getBigN();
//        
//        if (bp.isFileoutput())   
//            file.appendFile(name, number, testMap, correctionMap, mapSmallX, mapSmallN, bigX, bigN, selectedNodes, noClassificationsSet);
//    }
    
    public void calculateGoEnrichmentNew (HashSet selectedNodes, HashSet allNodes, String name, int number) {

    	HashMap testMap = null;
    	HashMap correctionMap = null;
    	HashMap mapSmallX = null;
    	HashMap mapSmallN = null;
    	HashMap mapBigX = null;
    	HashMap mapBigN = null;
    	HashSet noClassificationsSet = new HashSet() ;
    	int[] classifications = getClassificationsFromVector(selectedNodes, noClassificationsSet) ;

    	BingoAlgorithm algorithm = new BingoAlgorithm(bp);

    	CalculateTestTask test = algorithm.calculate_distribution();
    	//boolean success = TaskManager.executeTask(test, config);
    	test.run();
    	testMap = test.getTestMap();
    	CalculateCorrectionTask correction = algorithm.calculate_corrections(testMap);
    	if (correction != null) {
    		//success = TaskManager.executeTask(correction, config);
    		correction.run() ;
    		correctionMap = correction.getCorrectionMap();
    	}

    	//correctionMap = algorithm.calculate_corrections(testMap);

    	mapSmallX = test.getMapSmallX();
    	mapSmallN = test.getMapSmallN();
    	mapBigX = test.getMapBigX();
    	mapBigN = test.getMapBigN();

    	if (bp.isFileoutput())
    		file.appendFile(name,
    				number,
    				testMap,
    				correctionMap,
    				mapSmallX,
    				mapSmallN,
    				mapBigX,
    				mapBigN,
    				selectedNodes,
    				noClassificationsSet);
    } 
    
    public String [] ordenKeys(HashMap map) {
        
        String[] keySet = new String[map.keySet().size()];
        int count = 0 ;
        for(Object k: map.keySet()){
            keySet[count] = k.toString();
            count++;
        }
        
        for (int i = 1; i < keySet.length; i++) {
            int j = i;
            // get the first unsorted value ...
            String insert_key = keySet[i];
            double val = Double.parseDouble(map.get(keySet[i]).toString());
            // ... and insert it among the sorted
            while ((j > 0) && (val < Double.parseDouble(map.get(keySet[j-1]).toString()))) {
                keySet[j] = keySet[j - 1];
                j--;
            }
            // reinsert value
            keySet[j] = insert_key;
        }
        return keySet;
    }
}

