/* LeMoNe - Learning Module Network
 * 
 * Copyright (C) 2005-2008 VIB 
 * 
 * Written by Tom Michoel, Eric Bonnet, Anagha Joshi, Steven Maere
 *  
*/

package be.ugent.psb.ModuleNetwork;

import java.util.*;
import java.util.List;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.font.*;
import java.awt.geom.*;
import javax.imageio.*;
import javax.swing.*;
import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.impl.DenseDoubleMatrix2D;
import cern.colt.matrix.linalg.Algebra;
import be.ugent.psb.SetOperations.*;
import be.ugent.psb.GaneSh.*;
import cytoscape.data.annotation.OntologyTerm;
import net.sf.epsgraphics.*;

/**
 * Collection of genes with one or more possible regulation programs. 
 * The main building block in a {@link ModuleNetwork}. It extends 
 * JComponent to allow a graphical representation of the regulation
 * program(s).
 * 
 * @author tomic
 *
 */
public class Module extends JComponent {

	/**
	 * ModuleNetwork the module belongs to.
	 */
	public ModuleNetwork moduleNetwork;

	/**
	 * Numeric ID of module.
	 */
	public int number;

	/**
	 * Name of module (currently not used)
	 */
	public String name;

	/**
	 * List of genes in the module.
	 */ 
	public ArrayList<Gene> genes;

	/** 
	 * List of parents of the module.
	 */
	public ArrayList<Gene> parents;

	/**
	 * Map of all regulators (not just those appearing in a tree) to their weight.
	 */
	public HashMap<Gene, Double> regulatorWeights;
	
	/**
	 * Map of all parents occuring in {@link Module#hierarchicalTrees} to their weight,
	 * computed using the maximum weight in the testSplits of each node
	 */
	public HashMap<Gene, Double> regulatorWeightsMax;
	
	/**
	 * Map of all parents occuring in {@link Module#hierarchicalTrees} to their weight,
	 * computed using the sum of all weights in the testSplits of each node
	 */
	public HashMap<Gene, Double> regulatorWeightsSum;
	
	/**
	 * Map of all parents occuring in {@link Module#hierarchicalTrees} to their expected sign,
	 * computed using the sum of all weights in the testSplits of each node.
	 */
	public HashMap<Gene, Double> regulatorSign;
	
	/**
	 * Map of all parents occuring in {@link Module#hierarchicalTrees} to their left condition set,
	 * computed using the sum of all weights in the testSplits of each node. It gives for each condition
	 * the probability to belong to the condition set.
	 */
	public HashMap<Gene, ArrayList<Double>> regConditionsLeft;
	
	/**
	 * Map of all parents occuring in {@link Module#hierarchicalTrees} to their right condition set,
	 * computed using the sum of all weights in the testSplits of each node. It gives for each condition
	 * the probability to belong to the condition set.
	 */
	public HashMap<Gene, ArrayList<Double>> regConditionsRight;
	
	/**
	 * Map of all parents occuring in {@link Module#hierarchicalTrees} to their complete condition set,
	 * computed using the sum of all weights in the testSplits of each node. It gives for each condition
	 * the probability to belong to the condition set.
	 */
	public HashMap<Gene, ArrayList<Double>> regConditions;
	
	/**
	 * Map of all parents occuring in {@link Module#hierarchicalTrees} to their sign
	 */
	public HashMap<Gene, Integer> parentSigns;
	
	/**
	 * Map of all random parents occuring in {@link Module#hierarchicalTrees} to 
	 * their weight.
	 */
	public HashMap<Gene, Double> regulatorWeightsRandom;

	/**
	 * Hierarchical tree representing the regulation program. Used in 
	 * {@link ModuleNetwork#heuristicSearchMax(double, boolean, double, String[])} 
	 * when one tree per module is inferred.
	 */
	public TreeNode hierarchicalTree; 
	
	/**
	 * LEAF node to hold non-informative conditions (= conditions not in hierarchicalTree)
	 */
	public TreeNode nonInformCond;
	
	
	/**
	 * Different equally likely hierarchical trees. Based on different equally likely 
	 * clusterings of the experiments. Used in 
	 * {@link ModuleNetwork#gibbsSamplerExpts(int, int, int, int, double, boolean, boolean, double[])}.
	 */
	public ArrayList<TreeNode> hierarchicalTrees; 
	
	/**
	 * Map of each experiment to the leaf number it belongs to in each
	 * hierarchical tree
	 */
	public HashMap<Integer, ArrayList<Integer>> inExpCluster;
	
	/**
	 * Current best split for this module. Used in top-down search for regulation programs.
	 */
	public Split bestSplit;
	
	/**
	 * Module score. Computed by acting with {@link TreeNode#bayesianScore()} on
	 * {@link Module#hierarchicalTree}.
	 */
	public double moduleScore;

	/**
	 * Keep track if the list of genes has been changed compared to a previous situation.
	 * Used in {@link ModuleNetwork#heuristicSearchMax(double, boolean, double, String[])} to avoid
	 * rebuilding {@link Module#hierarchicalTree} if no genes were added/removed during
	 * {@link ModuleNetwork#moduleReassign(double)}.
	 */
	public boolean changed;
	
	/**
	 * GO
	 */
	
//	public BingoResults bingoResults;

	/**
	 *  GOsuper set to store GO "super" categories (highest term in the hierarchy) for this module.
	 */
	public LinkedHashSet<OntologyTerm> GOsuper;

	/**
	 *  GOall set to store all GO categories for this module. 
	 */
	public LinkedHashSet<OntologyTerm> GOall;

	/**
	 * ArrayList of top regulators for this module (post-processing data)
	 */
	public ArrayList<Gene> topRegulators;

	/**
	 * module mean gene expression value
	 */
	public double mean;
	
	/**
	 * module gene expression standard deviation value
	 */
	public double sigma;
	
	
	/**
	 * Empty constructor
	 */
	public Module() {
		this.changed = true;
		this.genes = new ArrayList<Gene>();
	    this.GOsuper = new LinkedHashSet<OntologyTerm>();
	    this.GOall = new LinkedHashSet<OntologyTerm>();
	    this.topRegulators = new ArrayList<Gene>();
	}

	/**
	 * Constructs initial module without regulation program.
	 * 
	 * @param M module network the module belongs to.
	 * @param number numeric ID of module.
	 * @param normalGammaPrior normal-gamma prior parameters used in the Bayesian score.
	 * 
	 */
	public Module(ModuleNetwork M, int number, double[] normalGammaPrior) {
		this.moduleNetwork = M;
		this.number = number;
		this.hierarchicalTree = new TreeNode(this, normalGammaPrior);
		for (int m =0; m<M.data[0].length; m++)
			this.hierarchicalTree.leafDistribution.condSet.add(m);
		this.changed = true;
		this.hierarchicalTrees = new ArrayList<TreeNode>();
		this.genes = new ArrayList<Gene>();
		this.parents = new ArrayList<Gene>();
		this.regulatorWeights = new HashMap<Gene,Double>();
		this.regulatorWeightsMax = new HashMap<Gene,Double>();
		this.regulatorWeightsRandom = new HashMap<Gene,Double>();
	    this.GOsuper = new LinkedHashSet<OntologyTerm>();
	    this.GOall = new LinkedHashSet<OntologyTerm>();
	    this.topRegulators = new ArrayList<Gene>();
	}

	// METHODS
	
	public void assignRegulatorsNoAcyclStoch (int level, double betaMax, int numRegAssign) {
    	System.out.println("\nAssigning regulators stochastically down to level " + level + " ...");
    	if (!this.genes.isEmpty()){
    		// check for multiple trees
    		if (this.hierarchicalTrees.isEmpty()){
    			this.hierarchicalTrees.add(this.hierarchicalTree);
    		} 
    		for (TreeNode root : this.hierarchicalTrees) {
    			// trim root below level
    			root.testLevel(level);
    			// assign regulators to remaining nodes
    			for (TreeNode node : root.getInternalNodes()){
    				node.findRegulatorBayesStoch(numRegAssign, this.moduleNetwork.regulatorSet, betaMax);
    				//node.testSplits = node.computeRegulatorSplitsBayes(this.moduleNetwork.regulatorSet, 20);
    				//node.testSplitsRandom = new ArrayList<Split>();
    			}
    		}
    		// set parents, default parameter ...
    		//this.setRegulatorWeightsMax(0, 0.5);
    		//this.setRegulatorWeightsSum(0, 0.5);
    	}
    	System.out.println("... done.");
    }
	
	/**
	 * 
	 */
	public void setInExpCluster(){
		// initialize
		this.inExpCluster = new HashMap<Integer, ArrayList<Integer>>();
		// map to find experiments by number
		//HashMap<Integer, Experiment> expMap = new HashMap<Integer, Experiment>();
		// fill
		for (int m=0; m<this.moduleNetwork.data[0].length; m++){
			this.inExpCluster.put(m, new ArrayList<Integer>());
		}
		
		for (TreeNode root : this.hierarchicalTrees){
			ArrayList<TreeNode> leaves = root.getLeafNodes();
			for (int k=0; k<leaves.size(); k++){
				for (Integer m : leaves.get(k).leafDistribution.condSet){
					this.inExpCluster.get(m).add(k);
				}
			}
		}
	}
	
	/**
	 * 
	 * @param file
	 */
	public void printExpClusters(String file){
		this.setInExpCluster();
		try{
    		File f = new File(file);
    		PrintWriter pw = new PrintWriter(f);
    		for (int e=0; e<this.moduleNetwork.data[0].length; e++){
    			String s = String.format("%s\t%d", "exp", e); //String.format("%.5s", gene.name) + "  \t" + gene.number;
    			for (int m : this.inExpCluster.get(e))
    				s += "\t" + m;
    			pw.println(s);
    		}
    		pw.close();
    	} catch (IOException e){
			System.out.println(e);
		}
	}
	
	public ArrayList<ArrayList<Integer>> readExpClusters (String file, int numCluster){
	       System.out.println("Reading experiment clusters ...");
	        int assigned = 0;
	        
	        ArrayList<ArrayList<Integer>> expclust = new ArrayList<ArrayList<Integer>>(numCluster);
	        for (int k=0; k<numCluster; k++)
	        	expclust.add(new ArrayList<Integer>());
	        
	        try {
	            Scanner clusterScanner = new Scanner(new File(file)).useDelimiter("\\n");
	            // walk through clusterfile, skip comment lines, check if counter starts with 0 or 1
	            int first = 1;
	            while (clusterScanner.hasNext()) {
	                Scanner line = new Scanner(clusterScanner.next().toString()).useDelimiter("\\s");
	                // read experiment number
	                int exp = line.nextInt();
	                int number = line.nextInt();
	                if (number == 0){
	                	first = 0;
	                }
	                expclust.get(number-first).add(exp);
	                assigned += 1;
	            }
	            System.out.println("... assigned " + assigned	+ " experiments to "
	                    + expclust.size() + " clusters.");
	            System.out.println();
	        } catch (FileNotFoundException e) {
	            System.out.println("Cluster file not found.");
	        }
	        
	        return expclust;
	}
	
	
	/**
	 * 
	 * @param numRuns
	 * @param numSteps
	 * @param burnIn
	 * @param sampleStep
	 * @param scoregain
	 * @param useBHCscore
	 */
	public void gibbsSampler(int numRuns, int numSteps, int burnIn, int sampleStep,
    		double scoregain, boolean useBHCscore){
		System.out.println("Gibbs sampling experiment partitions ...");
		this.hierarchicalTrees = new ArrayList<TreeNode>();
		HashSet<ArrayList<ArrayList<Integer>>> clusterList 
			= this.gibbsSamplerClustering(numRuns, numSteps, burnIn, sampleStep);
		for (ArrayList<ArrayList<Integer>> cluster : clusterList){
			this.hierarchicalTree = this.hierarchicalClusteringOrdered(this.partition2TreeNode(cluster), useBHCscore);
			this.hierarchicalTree.testScore(scoregain);
			this.hierarchicalTrees.add(this.hierarchicalTree);
		}
		System.out.println("... done");
	}
	
	/**
	 * Gibbs sampling to construct several equally likely
	 * partitions of the experiments for this module.
	 * 
	 * @see CRC
     * @param numRuns number of Gibbs sampler runs.
     * @param numSteps number of steps in each run of the Gibbs sampler.
     * @param burnIn number of intial steps before sampling starts.
     * @param sampleStep number of steps between successive samples.
     * 
     * @return Set of partitions. Each partition is a list of lists of integers
     * (referring to the experiment ID's) forming the sets of the partitions.
     * 
     * @author anjos, tomic
	 */
	public HashSet<ArrayList<ArrayList<Integer>>> gibbsSamplerClustering
		(int numRuns, int numSteps, int burnIn, int sampleStep) {

		// set to store different experiment clusterings
		HashSet<ArrayList<ArrayList<Integer>>> expClustering = new HashSet<ArrayList<ArrayList<Integer>>>();

		// get the data submatrix for this module
		DoubleMatrix2D data = new DenseDoubleMatrix2D(this.moduleNetwork.data);
		int[] rowList = new int[genes.size()];
		int i=0;
		for (Gene gene : genes)	{
			rowList[i]=gene.number;
			i++;
		}
		double[][] moduledata = Algebra.DEFAULT.transpose(data.viewSelection(rowList,null)).toArray();

		// cluster experiments using GaneSh
		for(int j=1; j<=numRuns; j++){
			GibbsSampler C = new GibbsSampler(moduledata);
			C.AssignRandomrowSinglecolumn();
			for(int k=1; k<=numSteps; k++){
				C.Cluster();
				if(k > burnIn && (numSteps - k) % sampleStep == 0){
					// list of experiment clusters 
					ArrayList<ArrayList<Integer>> expClusters = new ArrayList<ArrayList<Integer>>();
					for(Cluster clust : C.ClusterSet){
						ArrayList<Integer> expList 
							= new ArrayList<Integer>(clust.RowSet.size());
						for(Integer in : clust.RowSet){
							expList.add(in);
						}
						expClusters.add(expList);
					}
					expClustering.add(expClusters);
				}
			}
		}
		return(expClustering);
	} 

	/**
	 * Converts a partition of the experiments to a list of {@link TreeNode}s. 
	 * For each set of experiments, creates a TreeNode with this set as its condition set. 
	 * This list of TreeNodes can be used as input for 
	 * {@link Module#hierarchicalClustering(List, boolean)}.
	 * 
	 * @param clusters partition of experiments, list of lists of integers.
	 * @return List of {@link TreeNode}s.
	 * 
	 * @author tomic
	 *   
	 */
	public List<TreeNode> partition2TreeNode (ArrayList<ArrayList<Integer>> clusters) {
		List<TreeNode> treeList = new ArrayList<TreeNode>();
		
		for (ArrayList<Integer> list : clusters){
			TreeNode node = new TreeNode(this, this.hierarchicalTree.leafDistribution.normalGammaPrior);
			node.leafDistribution.condSet = list;
			node.statistics();
			node.leafDistribution.bayesianScore();
			treeList.add(node);
		}
		
		return treeList;
	}
	
	
	/**
	 * Takes the output of {@link Module#gibbsSamplerClustering(int, int, int, int)} and creates a list
	 * of condition sets which are clustered together in each Gibbs sample and have a minimal size.
	 * 
	 * @param partitionList
	 * @param minSize
	 * @return
	 * 
	 * @author tomic
	 */
	public ArrayList<ArrayList<Integer>> makeConsensusCondSets (HashSet<ArrayList<ArrayList<Integer>>> partitionList, int minSize){
		// create list of intersections of all partitions
		Iterator<ArrayList<ArrayList<Integer>>> it = partitionList.iterator();
		ArrayList<ArrayList<Integer>> list = it.next();
		while (it.hasNext()){
			list = SetOperation.prodPart(list, it.next());
		}
		// find intersections that are too small
		ArrayList<ArrayList<Integer>> remove = new ArrayList<ArrayList<Integer>>();
		for (ArrayList<Integer> l : list)
			if (l.size() < minSize)
				remove.add(l);
		for (ArrayList<Integer> l : remove)
			list.remove(l);
		// done
		return list;
	}
	
	/**
	 * Takes the leaf nodes of all trees in {@link Module#hierarchicalTrees} and creates a list
	 * of condition sets which are clustered together in each tree and have a minimal size.
	 * 
	 * @param partitionList
	 * @param minSize
	 * @return
	 * 
	 * @author tomic
	 */
	public ArrayList<ArrayList<Integer>> makeConsensusCondSetsFromTrees (int minSize){
		// create partition list
		HashSet<ArrayList<ArrayList<Integer>>> partitionList = new HashSet<ArrayList<ArrayList<Integer>>> ();
		for (TreeNode root : this.hierarchicalTrees){
			partitionList.add(root.condSets());
		}
		// create list of intersections of all partitions
		Iterator<ArrayList<ArrayList<Integer>>> it = partitionList.iterator();
		ArrayList<ArrayList<Integer>> list = it.next();
		while (it.hasNext()){
			list = SetOperation.prodPart(list, it.next());
		}
		// find intersections that are too small
		ArrayList<ArrayList<Integer>> remove = new ArrayList<ArrayList<Integer>>();
		for (ArrayList<Integer> l : list)
			if (l.size() < minSize)
				remove.add(l);
		for (ArrayList<Integer> l : remove)
			list.remove(l);
		// done
		return list;
	}
	
	/**
	 * Makes a consensus tree from {@link Module#hierarchicalTrees}, stores the result in {@link Module#hierarchicalTree},
	 * and empties {@link Module#hierarchicalTrees}.
	 * @param minSize
	 * 
	 * @author tomic
	 */
	public void makeConsensusTree (int minSize){
		ArrayList<ArrayList<Integer>> condsets =  this.makeConsensusCondSetsFromTrees(minSize);
		if (condsets.size() > 0){
			// cluster hierarchically
			this.hierarchicalTree = this.hierarchicalClusteringOrdered(this.partition2TreeNode(condsets), true);
		} else {
			this.hierarchicalTree.nodeStatus = "leaf";
			this.hierarchicalTree.leftChild = null;
			this.hierarchicalTree.rightChild = null;
		}
		// empty trees
		this.hierarchicalTrees = new ArrayList<TreeNode>();
	}
	
	/**
	 * Hierarchical clustering of experiments (or experiment clusters).
	 * This builds a full hierarchical clustering tree of the data in the
	 * module, which can then be converted in a regulation program tree.
	 * 
	 * Usage of {@link TreeNode} and {@link LeafDistribution} is somewhat 
	 * non-standard:
	 * <ul> 
	 * <li>{@link TreeNode#testGene} and {@link TreeNode#splitValue} are not set on 
	 * internal nodes.</li>
	 * <li>Internal nodes have a {@link LeafDistribution} as well, containing 
	 * the statistics of all leaves (=experiments) below it.</li> 
 	 * </ul>
 	 * 
 	 * The hierarchical clustering proceeds by merging in each step the pair of 
 	 * nodes with the smallest {@link TreeNode#mergeScore}. This can be either
 	 * the score difference between the merged nodes and the sum of scores of each 
 	 * node, or the score based on Bayesian Hierarchical Clustering which also 
 	 * takes into account the other partitions below the pair of nodes.
	 * 
	 * @param treeList list of TreeNodes to be clustered. This can be each 
	 * experiment in an individual leaf node such as returned by {@link Module#initTreeList()} or
	 * {@link Module#initTreeList(ArrayList)}, 
	 * or a set of experiment clusters such as returned by 
	 * {@link Module#partition2TreeNode(ArrayList)}.
	 * @param useBHCscore use merge score from Bayesian Hierarchical Clustering or 
     * simple score difference when hierarchically clustering experiments.
	 * @return root node of the hierarchical tree.
	 * 
	 * @author tomic
	 * 
	 */
	public TreeNode hierarchicalClustering(List<TreeNode> treeList, boolean useBHCscore) {
		int numNode = treeList.size();
		double[] normalGammaPrior = treeList.get(0).leafDistribution.normalGammaPrior;
		List<TreeNode> pairTreeList = new ArrayList<TreeNode>(numNode*(numNode-1)/2);
		
		// initialize list of pairwise merges
		for (int i = 0; i < numNode; i++){
			for (int j = 0; j < i; j++) {
				TreeNode tree = new TreeNode(this, normalGammaPrior);
				tree.nodeStatus = "internal";
				// child nodes, order by means of normal distribution at nodes
				if (treeList.get(i).leafDistribution.gaussianParam[0] < 
						treeList.get(j).leafDistribution.gaussianParam[0]){
					tree.leftChild = treeList.get(i);
					tree.rightChild = treeList.get(j);
				} else {
					tree.leftChild = treeList.get(j);
					tree.rightChild = treeList.get(i);
				}
				// statistics of merged node
				for (int s = 0; s < 3; s++)
					tree.leafDistribution.statistics[s] = tree.leftChild.leafDistribution.statistics[s]
							+ tree.rightChild.leafDistribution.statistics[s];
				// union of condition sets
				tree.leafDistribution.condSet = SetOperation.union(tree.leftChild.leafDistribution.condSet, 
						tree.rightChild.leafDistribution.condSet);
				// compute score
				tree.leafDistribution.bayesianScore();
				tree.setMergeScore(useBHCscore);
				// add to list
				pairTreeList.add(tree);
			}
		}

		// merge trees in treeList until one tree is left
		TreeNode bestTree = new TreeNode(this, normalGammaPrior);
		bestTree.mergeScore = 1.0;
		while (treeList.size() > 1) {
			// find the pair tree with the highest posterior probability
			//Collections.sort(pairTreeList);
			bestTree = Collections.max(pairTreeList);
			// set bestTree as parent of its children
			bestTree.leftChild.parentNode = bestTree;
			bestTree.rightChild.parentNode = bestTree;
			// remove children of bestTree from treeList; bestTree will be added at the end
			treeList.remove(bestTree.leftChild);
			treeList.remove(bestTree.rightChild);
			// loop over the pair list and remove trees whose condSet intersects
			// with the condSet of the highest scoring tree, including the highest one
			// itself
			Iterator<TreeNode> it = pairTreeList.iterator();
			while (it.hasNext()) {
				TreeNode tree = it.next();
				ArrayList<Integer> intersect 
					= SetOperation.intersection(tree.leafDistribution.condSet, 
							bestTree.leafDistribution.condSet);
				if (!intersect.isEmpty())
					it.remove();
			}
			// add new trees which merge the bestTree with the other trees in
			// treeList;
			Iterator<TreeNode> it2 = treeList.iterator();
			while (it2.hasNext()) {
				TreeNode ittree = it2.next();
				TreeNode tree = new TreeNode(this, normalGammaPrior);
				tree.nodeStatus = "internal";
				// child nodes, order means of normal distributions at the nodes
				if (ittree.leafDistribution.gaussianParam[0] < 
						bestTree.leafDistribution.gaussianParam[0]){
					tree.leftChild = ittree;
					tree.rightChild = bestTree;
				} else {
					tree.leftChild = bestTree;
					tree.rightChild = ittree;
				}
				// statistics of merged node
				for (int s = 0; s < 3; s++)
					tree.leafDistribution.statistics[s] = tree.leftChild.leafDistribution.statistics[s]
							+ tree.rightChild.leafDistribution.statistics[s];
				// union of condition sets
				tree.leafDistribution.condSet = SetOperation.union(tree.leftChild.leafDistribution.condSet, 
						tree.rightChild.leafDistribution.condSet);
				// compute scores
				tree.leafDistribution.bayesianScore();
				tree.setMergeScore(useBHCscore);
				// add to list
				pairTreeList.add(tree);
			}
			// add bestTree again to treeList
			treeList.add(bestTree);
		}
		// 1 tree left, this is the one
		return treeList.get(0);
	}
	
	/**
	 * Hierarchical clustering of a set of leaf nodes. Similar to {@link Module#hierarchicalClustering(List, boolean)},
	 * but this method orders the nodes by their mean expression level and only allows merges
	 * between neighboring nodes. This is faster and prevents the 'wrong' linking that is sometimes seen in
	 * {@link Module#hierarchicalClustering(List, boolean)}.
	 * 
	 * @param treeList
	 * @param useBHCscore
	 * @return
	 * 
	 * @author tomic
	 */
	public TreeNode hierarchicalClusteringOrdered(List<TreeNode> treeList, boolean useBHCscore) {
		double[] normalGammaPrior = treeList.get(0).leafDistribution.normalGammaPrior;
		// list of possible merges
		List<TreeNode> pairList = this.pairTreeList(treeList, useBHCscore);
		// merge trees in treeList until one tree is left
		TreeNode bestTree = new TreeNode(this, normalGammaPrior);
		bestTree.mergeScore = 1.0;
		while (treeList.size() > 1) {
			// find the pair tree with the highest posterior probability
			bestTree = Collections.max(pairList);
			// set bestTree as parent of its children
			bestTree.leftChild.parentNode = bestTree;
			bestTree.rightChild.parentNode = bestTree;
			// remove children of bestTree from treeList and add bestTree
			treeList.remove(bestTree.leftChild);
			treeList.remove(bestTree.rightChild);
			treeList.add(bestTree);
			// sort and pair again
			pairList = this.pairTreeList(treeList, useBHCscore);
		}
		// 1 tree left, this is the one
		return treeList.get(0);
	}
	
	/**
	 * Creates from a list of tree nodes a new list of tree nodes by merging
	 * neighboring pairs (neighbor in the sense of data mean). 
	 * 
	 * @param treeList
	 * @param useBHCscore
	 * @return
	 * 
	 * @author tomic
	 */
	public List<TreeNode> pairTreeList(List<TreeNode> treeList, boolean useBHCscore){
		double[] normalGammaPrior = treeList.get(0).leafDistribution.normalGammaPrior;
		// sort treeList by increasing mean
		Comparator mc = new MeanComparator();
		Collections.sort(treeList, mc);
		// list of merges between neighboring pairs
		List<TreeNode> pairList = new ArrayList<TreeNode>(treeList.size()-1);
		for (int i = 0; i < treeList.size()-1; i++){
			TreeNode tree = new TreeNode(this, normalGammaPrior);
			// attach children
			tree.nodeStatus = "internal";
			// child nodes	
			tree.leftChild = treeList.get(i);
			tree.rightChild = treeList.get(i+1);
			// statistics of merged node
			for (int s = 0; s < 3; s++)
				tree.leafDistribution.statistics[s] = tree.leftChild.leafDistribution.statistics[s]
						+ tree.rightChild.leafDistribution.statistics[s];
			// union of condition sets
			tree.leafDistribution.condSet = SetOperation.union(tree.leftChild.leafDistribution.condSet, 
					tree.rightChild.leafDistribution.condSet);
			// compute score
			tree.leafDistribution.bayesianScore();
			tree.setMergeScore(useBHCscore);
			// dummy regulator
			tree.regulationSplit = new Split();
			tree.testSplits = new ArrayList<Split>();
			// add to list
			pairList.add(tree);
		}
		return pairList;
	}
	
	/**
	 * From a list of experiment clusters (in the form of leaf nodes), find a
	 * subset which form mutually significant splits. The siginificance test
	 * is whatever is implemented in 
	 * {@link Module#compareSignificance(TreeNode, TreeNode)} 
	 * The idea is to apply this method
	 * to the list of leaves obtained from {@link Module#makeConsensusTree(int)}.
	 * 
	 * @param treeList
	 * @param fact multiplication factor in significance test
	 * @return
	 * 
	 * @author tomic
	 */
	public TreeNode findSignificantSplits (List<TreeNode> treeList, double fact){
		int numTree = treeList.size();
		// help tree list to store selected nodes from treeList
		List<TreeNode> tmpList = new ArrayList<TreeNode>(); 
		// sort treeList by increasing mean
		Comparator mc = new MeanComparator();
		Collections.sort(treeList, mc);
		
		if (treeList.size()==1)
			return treeList.get(0);
		else {
			// compare first and last
			boolean init = this.compareSignificance(treeList.get(0), treeList.get(numTree-1), fact);
			if (!init){ // put everything in 1 leaf
				TreeNode tree = this.hierarchicalClusteringOrdered(treeList, true);
				tree.testLevel(-1);
				return tree;
			} else {// we have something
				tmpList.add(treeList.get(0));
				tmpList.add(treeList.get(numTree-1));
				treeList.remove(numTree-1);
				treeList.remove(0);
				if (!treeList.isEmpty()){// recursion
					TreeNode tree2 = this.findSignificantSplits(treeList, fact);
					// check if we can add tree2
					boolean add = true;
					for (TreeNode node : tmpList){
						add = add && this.compareSignificance(node, tree2, fact);
					}
					if (add)
						tmpList.add(tree2);
				}
				// link chosen nodes into 1 tree
				TreeNode tree = this.hierarchicalClusteringOrdered(tmpList, true);
				return tree;
			}
		}
	}
	
	public void expandSignificantSplits(double fact){
		this.setNonInformCond();
		if (!this.nonInformCond.leafDistribution.condSet.isEmpty()){
			int num = 0;
			ArrayList<TreeNode> leaves = this.hierarchicalTree.getLeafNodes();
			for (int m : this.nonInformCond.leafDistribution.condSet){
				double mean = 0;
				for (Gene gene : this.genes){
					mean += this.moduleNetwork.data[gene.number][m];
				}
				mean /= this.genes.size();
				// add m to a leaf if it is within (fact x standard deviation) of the mean;
				// because of significant splits this can happen with at most 1 leaf, if "fact"
				// has the same value as used to determine significance
				for (TreeNode leaf : leaves){
					if (Math.abs(leaf.leafDistribution.gaussianParam[0] - mean) < fact*leaf.leafDistribution.gaussianParam[1]){
						// add m
						leaf.leafDistribution.condSet.add(m);
						// update statistics
						leaf.statistics();
						// update score and normal dist param
						leaf.leafDistribution.bayesianScore();
						// done
						num += 1;
						break;
					}
				}
			}
			this.setNonInformCond();
//			System.out.println(num + "/" + this.nonInformCond.leafDistribution.condSet.size() + " conditions added");
		}
	}
	
	/**
	 * Checks if the normal distributions for two nodes are significantly separated, i.e.,
	 * if (mu1 - mu2) > fact*(sigma1 + sigma2)
	 * 
	 * @param node1
	 * @param node2
	 * @param fact
	 * @return
	 * 
	 * @author tomic
	 */
	public boolean compareSignificance(TreeNode node1, TreeNode node2, double fact){
		Double mu1 = node1.leafDistribution.gaussianParam[0];
		Double sigma1 = node1.leafDistribution.gaussianParam[1];
		Double mu2 = node2.leafDistribution.gaussianParam[0];
		Double sigma2 = node2.leafDistribution.gaussianParam[1];
		if (Math.abs(mu1-mu2) > fact*(sigma1+sigma2))
			return true;
		else
			return false;
	}
	
	
	
	/**
	 * Putd each experiment in its own leaf node. This list of TreeNodes can be
	 * used as input for {@link Module#hierarchicalClustering(List, boolean)}.
	 * 
	 * @author tomic
	 */
	public List<TreeNode> initTreeList () {
		double[][] data = this.moduleNetwork.data;
		List<TreeNode> treeList = new ArrayList<TreeNode>(data[0].length);
		for (int i=0; i<data[0].length; i++) {
			TreeNode tree = new TreeNode(this, this.moduleNetwork.normalGammaPrior);
			tree.nodeStatus = "leaf";
			tree.leafDistribution.condSet.add(i);
			for (Gene gene : this.genes)
				if (!Double.isNaN(data[gene.number][i])) {
					tree.leafDistribution.statistics[0]++;
					tree.leafDistribution.statistics[1] += data[gene.number][i];
					tree.leafDistribution.statistics[2] += Math.pow(data[gene.number][i], 2);
				}
			tree.leafDistribution.bayesianScore();
			tree.mergeScore = 1.0;
			treeList.add(tree);
		}
		return treeList;
	}
	
	/**
	 * Puts each experiment of a selected list in its own leaf node. This list of TreeNodes 
	 * can be used as input for {@link Module#hierarchicalClustering(List, boolean)}.
	 * 
	 * @param condSet for each experiment in this list a leaf node is created.
	 * 
	 * @author tomic
	 * 
	 */
	public List<TreeNode> initTreeList (ArrayList<Integer> condSet) {
		List<TreeNode> treeList = new ArrayList<TreeNode>(condSet.size());
		double[][] data = this.moduleNetwork.data;
		for (int i : condSet) {
			TreeNode tree = new TreeNode(this, this.moduleNetwork.normalGammaPrior);
			tree.nodeStatus = "leaf";
			tree.leafDistribution.condSet.add(i);
			for (Gene gene : this.genes)
				if (!Double.isNaN(data[gene.number][i])) {
					tree.leafDistribution.statistics[0]++;
					tree.leafDistribution.statistics[1] += data[gene.number][i];
					tree.leafDistribution.statistics[2] += Math.pow(data[gene.number][i], 2);
				}
			tree.leafDistribution.bayesianScore();
			tree.mergeScore = 1.0;
			treeList.add(tree);
		}
		return treeList;
	}

	
	public void setRegulatorSign (){
		// initialize
		int numCond = this.moduleNetwork.data[0].length;
		this.regulatorSign = new HashMap<Gene,Double>();
		for (Gene reg : this.regulatorWeights.keySet())
			this.regulatorSign.put(reg, 0.0);
		// compute
		if (this.hierarchicalTrees.isEmpty())
			this.hierarchicalTrees.add(this.hierarchicalTree);
		for (TreeNode regTree: this.hierarchicalTrees) {
			for (TreeNode node : regTree.getInternalNodes()){
				double fact = Math.pow((double)node.leafDistribution.condSet.size()/(double)this.moduleNetwork.data[0].length,1);
				for (Split splt : node.testSplits){
					double pIn = this.regulatorSign.get(splt.regulator);
					double pAdd = splt.sign * Math.exp(splt.regulatorScore/node.leafDistribution.condSet.size())*fact
						/this.regulatorWeights.get(splt.regulator);
					this.regulatorSign.put(splt.regulator, pIn + pAdd);
				}
			}
		}
	}
	
	
	public void setRegConditions (){
		// initialize
		int numCond = this.moduleNetwork.data[0].length;
		this.regConditionsLeft = new HashMap<Gene, ArrayList<Double>>();
		this.regConditionsRight = new HashMap<Gene, ArrayList<Double>>();
		this.regConditions = new HashMap<Gene, ArrayList<Double>>();
		for (Gene reg : this.regulatorWeights.keySet()){
			this.regConditionsLeft.put(reg, new ArrayList<Double>(numCond));
			this.regConditionsRight.put(reg, new ArrayList<Double>(numCond));
			this.regConditions.put(reg, new ArrayList<Double>(numCond));
			for (int m=0; m<numCond; m++){
				this.regConditionsLeft.get(reg).add(0.0);
				this.regConditionsRight.get(reg).add(0.0);
				this.regConditions.get(reg).add(0.0);
			}
		}
		// compute
		if (this.hierarchicalTrees.isEmpty())
			this.hierarchicalTrees.add(this.hierarchicalTree);
		for (TreeNode regTree: this.hierarchicalTrees) {
			for (TreeNode node : regTree.getInternalNodes()){
				double fact = Math.pow((double)node.leafDistribution.condSet.size()/(double)this.moduleNetwork.data[0].length,1);
				for (Split splt : node.testSplits){
					for (int m : node.leftChild.leafDistribution.condSet){
						double pIn = this.regConditionsLeft.get(splt.regulator).get(m);
						double pAdd = Math.exp(splt.regulatorScore/node.leafDistribution.condSet.size())*fact
							/this.regulatorWeights.get(splt.regulator);
						this.regConditionsLeft.get(splt.regulator).set(m, pIn + pAdd);
						this.regConditions.get(splt.regulator).set(m, pIn + pAdd);
					}
					for (int m : node.rightChild.leafDistribution.condSet){
						double pIn = this.regConditionsRight.get(splt.regulator).get(m);
						double pAdd = Math.exp(splt.regulatorScore/node.leafDistribution.condSet.size())*fact
							/this.regulatorWeights.get(splt.regulator);
						this.regConditionsRight.get(splt.regulator).set(m, pIn + pAdd);
						this.regConditions.get(splt.regulator).set(m, pIn + pAdd);
					}
				}
			}
		}
	}
	/**
	 * Computes the regulator weights. Sums regulator weights at
	 * all nodes.
	 *
	 * @author tomic
	 */
	public void setRegulatorWeights(){
		this.regulatorWeights = new HashMap<Gene, Double>();
		if (this.hierarchicalTrees.isEmpty())
			this.hierarchicalTrees.add(this.hierarchicalTree);
		for (TreeNode regTree: this.hierarchicalTrees) {
			for (TreeNode node : regTree.getInternalNodes()){
				double fact = Math.pow((double)node.leafDistribution.condSet.size()/(double)this.moduleNetwork.data[0].length,1);
				for (Split splt : node.testSplits){
					if (!this.regulatorWeights.containsKey(splt.regulator))
						this.regulatorWeights.put(splt.regulator, 0.0);
					double w = this.regulatorWeights.get(splt.regulator)
						+ Math.exp(splt.regulatorScore/node.leafDistribution.condSet.size())*fact;
					this.regulatorWeights.put(splt.regulator, w);
				}
			}
		}
	}
	
//	public void setRegulatorWeightsNew(){
//		this.regulatorWeights = new HashMap<Gene, Double>();
//		if (this.hierarchicalTrees.isEmpty())
//			this.hierarchicalTrees.add(this.hierarchicalTree);
//		// tree normalization
//		double zTree = (double)this.hierarchicalTrees.size();
//		for (TreeNode regTree: this.hierarchicalTrees) {
//			// first compute node normalization
//			double zNode = 0.0;
//			for (TreeNode node : regTree.getInternalNodes())
//				zNode += (double)node.leafDistribution.condSet.size()/(double)this.moduleNetwork.data[0].length;
//			// next compute probabilities
//			for (TreeNode node : regTree.getInternalNodes()){
//				double wNode = ((double)node.leafDistribution.condSet.size()/(double)this.moduleNetwork.data[0].length)/(zNode*zTree);
//				node.setRegulatorWeightsNew();
//				for (Gene reg : node.regulatorWeights.keySet()){
//					if (!this.regulatorWeights.containsKey(reg))
//						this.regulatorWeights.put(reg, 0.0);
//					double w = this.regulatorWeights.get(reg) + wNode*node.regulatorWeights.get(reg);
//					this.regulatorWeights.put(reg, w);
//				}
//			}
//		}
//	}
	

	/**
	 * Computes the weight of each regulator at the nodes of all
	 * hierarchical trees for this module. The weight is given by
	 * the average probability of a correct prediction, i.e.,
	 * w = exp(regulatorScore/node.numConditions), summed over all
	 * occurrences of a regulator in the set of regulation trees.
	 *
	 * @param downWeightLevel reduce weights of assignments below this tree level
	 * @param downWeightFact multiplication factor by which to reduce weights
	 * @deprecated use setRegulatorWeights instead
	 * @author tomic
	 */
	public void setRegulatorWeightsSum(int downWeightLevel, double downWeightFact) {
		this.regulatorWeightsSum = new HashMap<Gene, Double>();
		for (TreeNode regTree: this.hierarchicalTrees) {
			for (TreeNode node : regTree.getInternalNodes()){
				double fact = 1.0;
//				if (node.treeLevel() > downWeightLevel)
//					fact = downWeightFact;
				//double fact = (double)node.leafDistribution.condSet.size()/(double)this.moduleNetwork.data[0].length;
				for (Split splt : node.testSplits){
					if (!this.regulatorWeightsSum.containsKey(splt.regulator))
						this.regulatorWeightsSum.put(splt.regulator, 0.0);
					double w = this.regulatorWeightsSum.get(splt.regulator)
						+ Math.exp(splt.regulatorScore/node.leafDistribution.condSet.size())*fact;
					this.regulatorWeightsSum.put(splt.regulator, w);
				}
			}
		}
	}
	
	
	/**
	 * Computes the weight of each regulator at the nodes of all
	 * hierarchical trees for this module. The weight is given by
	 * the average probability of a correct prediction, i.e.,
	 * w = exp(regulatorScore/node.numConditions), maximized over all
	 * occurrences of a regulator in the set of regulation trees.
	 *
	 * @param downWeightLevel reduce weights of assignments below this tree level
	 * @param downWeightFact multiplication factor by which to reduce weights
	 * @deprecated use setRegulatorWeights instead
	 * @author tomic
	 */
	public void setRegulatorWeightsMax(int downWeightLevel, double downWeightFact){
		this.regulatorWeightsMax = new HashMap<Gene, Double>();
		for (TreeNode regTree: this.hierarchicalTrees) {
			for (TreeNode node : regTree.getInternalNodes()){
//				double fact = 1.0;
//				if (node.treeLevel() > downWeightLevel)
//					fact = downWeightFact;
				double fact = (double)node.leafDistribution.condSet.size()/(double)this.moduleNetwork.data[0].length;
				//double randScore = -Math.log(2.0)*numCond;
				for (Split splt : node.testSplits){
					//double score = splt.regulatorScore - randScore;
					double score = Math.exp(splt.regulatorScore/node.leafDistribution.condSet.size())*fact;
					if (!this.regulatorWeightsMax.containsKey(splt.regulator))
						this.regulatorWeightsMax.put(splt.regulator, score);
					else{
						double w = this.regulatorWeightsMax.get(splt.regulator);
						if (score > w)
							this.regulatorWeightsMax.put(splt.regulator, score);
					}
				}
			}
		}
	}
	
	

	
	/**
	 * Computes the weight of each random regulator at the nodes of all
	 * hierarchical trees for this module. The weight is given by
	 * the average probability of a correct prediction, i.e.,
	 * w = exp(regulatorScore/node.numConditions), summed over all
	 * occurrences of a random regulator in the set of regulation trees.
	 *
	 * @author tomic
	 */
	public void setRegulatorWeightsRandom() {
		this.regulatorWeightsRandom = new HashMap<Gene, Double>();
		for (TreeNode regTree: this.hierarchicalTrees) {
			for (TreeNode node : regTree.getInternalNodes()){
				double fact = (double)node.leafDistribution.condSet.size()/(double)this.moduleNetwork.data[0].length;
				for (Split splt : node.testSplitsRandom){
					if (!this.regulatorWeightsRandom.containsKey(splt.regulator))
						this.regulatorWeightsRandom.put(splt.regulator, 0.0);
					double w = this.regulatorWeightsRandom.get(splt.regulator)
						+ Math.exp(splt.regulatorScore/node.leafDistribution.condSet.size())*fact;
					this.regulatorWeightsRandom.put(splt.regulator, w);
				}
			}
		}
	}
	
	/**
	 * Computes the weight of each random regulator at the nodes of all
	 * hierarchical trees for this module. The weight is given by
	 * the average probability of a correct prediction, i.e.,
	 * w = exp(regulatorScore/node.numConditions), summed over all
	 * occurrences of a random regulator in the set of regulation trees.
	 *
	 * @param downWeightLevel reduce weights of assignments below this tree level
	 * @param downWeightFact multiplication factor by which to reduce weights
	 * @deprecated use setRegulatorWeightsRandom instead
	 * @author tomic
	 */
	public void setParentWeightsRandomSum(int downWeightLevel, double downWeightFact) {
		this.regulatorWeightsRandom = new HashMap<Gene, Double>();
		for (TreeNode regTree: this.hierarchicalTrees) {
			for (TreeNode node : regTree.getInternalNodes()){
				double fact = 1.0;
				if (node.treeLevel() > downWeightLevel)
					fact = downWeightFact;
				for (Split splt : node.testSplitsRandom){
					if (!this.regulatorWeightsRandom.containsKey(splt.regulator))
						this.regulatorWeightsRandom.put(splt.regulator, 0.0);
					double w = this.regulatorWeightsRandom.get(splt.regulator)
						+ Math.exp(splt.regulatorScore/node.leafDistribution.condSet.size())*fact;
					this.regulatorWeightsRandom.put(splt.regulator, w);
				}
			}
		}
	}
	
	/**
	 * Sets the sign (activator/repressor) for each regulator at the nodes of all
	 * hierarchical trees for this module.
	 *
	 * @return false if a regulator occurs with different sign at different nodes
	 * 
	 * @author tomic
	 */
	public boolean setParentSigns(){
		boolean consistent = true;
		this.parentSigns = new HashMap<Gene, Integer>();
		for (TreeNode regTree: this.hierarchicalTrees) {
			for (TreeNode node : regTree.getInternalNodes()){
				for (Split splt : node.testSplits){
					splt.setSumData();
					splt.computeSign();
					if (!this.parentSigns.containsKey(splt.regulator))
						this.parentSigns.put(splt.regulator, splt.sign);
					else // check consistency
						if (splt.sign != this.parentSigns.get(splt.regulator))
							consistent = false;
				}
			}
		}
		return consistent;
	}

	/** 
	 * Sets the leaf node holding the non-informative conditions
	 * 
	 * @author tomic
	 */
	public void setNonInformCond(){
		this.nonInformCond = new TreeNode(this, this.moduleNetwork.normalGammaPrior);
		for (int m=0; m<this.moduleNetwork.data[0].length; m++)
			if (!this.hierarchicalTree.leafDistribution.condSet.contains(m))
				this.nonInformCond.leafDistribution.condSet.add(m);
		this.nonInformCond.statistics();
	}
	
	/**
	 * Validates the regulation program in hierarchicalTree on the conditions
	 * in nonInformCond
	 * 
	 * @author tomic
	 */
	public double validateRegProgram(){
		this.setNonInformCond();
		double p = 0.0;
		EnsembleCalculator calc = new EnsembleCalculator(this.moduleNetwork);
		for (int m : this.nonInformCond.leafDistribution.condSet){
			double[] xExp = calc.getDataColumn(m);
			for (Gene gene : this.genes){
				double pi = 0.0;
				for (TreeNode leaf : this.hierarchicalTree.getLeafNodes()){
					double w = leaf.pathWeight(xExp);
					pi += w * leaf.leafDistribution.evaluate(xExp[gene.number]);
				}
				p += Math.log(pi);
			}
		}
		return Math.exp(p/((double)this.nonInformCond.leafDistribution.condSet.size()*this.genes.size()));
	}
	
	/**
	 * Draws a figure of the current regulation program for this module.
	 * This method draws the expression values of the genes in the module, 
	 * partitioned according to the leaves in {@link Module#hierarchicalTree},
	 * together with the hierarchical tree. At each node of the tree it draws
	 * the expression values for the regulator at that node for the set of 
	 * experiments that is splitted by that node.
	 * 
	 * @param filename name of file to store figure
	 * 
	 * @author tomic, stmae
	 * 
	 */
	public void draw(String filename) {
		// scale factor
		int sc = 4;
		// coordinates of origin
		int x0 = 10;
		int y0 = 10;
		// tree step height
		int h = 18;
		// figure width
		int width = 3 * x0 + this.hierarchicalTree.leafDistribution.condSet.size();
		int height;
		int x = x0;
		int ybase = y0;

		if (this.hierarchicalTree.nodeStatus.equals("leaf")){ // no tree
			height = 2 * y0 + this.genes.size()*2;
		}	
		else {
			// coordinates of root
			x = x0 + this.hierarchicalTree.leftChild.leafDistribution.condSet.size();
			ybase = y0 + (h + 4) * (this.hierarchicalTree.treeDepth() + 1);
			height = ybase + this.genes.size()*2 + y0;
		}
		BufferedImage image = new BufferedImage(sc * width, sc * height, BufferedImage.TYPE_INT_BGR);
		Graphics2D g2;
		g2 = image.createGraphics();
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.scale(sc, sc);
		g2.setStroke(new BasicStroke(0.5F));
		if (this.hierarchicalTree.nodeStatus.equals("leaf")) { // no tree
			Collections.sort(this.hierarchicalTree.leafDistribution.condSet);
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			// draw data
			draw1leaf(x0, y0, g2);
		} else {
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			drawnode(this.hierarchicalTree, x, y0, h, ybase, g2);
		}
		File outFile = new File(filename);
		try {
			ImageIO.write(image, "png", outFile);
		} catch (IOException e) {
			return;
		}
	}
	
	/**
	 * Draws a figure of the current regulation program for this module with extra information.
	 * This method draws the expression values of the genes in the module, 
	 * partitioned according to the leaves in {@link Module#hierarchicalTree},
	 * together with the hierarchical tree. At each node of the tree it draws
	 * the expression values for the regulator at that node for the set of 
	 * experiments that is splitted by that node.
	 * 
	 * @param filename name of file to store figure
	 * 
	 * @author tomic, stmae
	 * 
	 */
	public void drawExtra(String filename, ArrayList<Gene> geneSet) {
		// scale factor
		int sc = 4;
		// coordinates of origin
		int x0 = 10;
		int y0 = 10;
		// tree step height
		int h = 18;
		// figure width
		this.setNonInformCond();
		if (this.hierarchicalTree.nodeStatus.equals("internal")){
			ExpComparator ec = new ExpComparator(this.moduleNetwork, this.hierarchicalTree.regulationSplit.regulator);
			Collections.sort(this.nonInformCond.leafDistribution.condSet, ec);
			for (TreeNode node : this.hierarchicalTree.getInternalNodes())
				Collections.sort(node.leafDistribution.condSet, ec);
			for (TreeNode node : this.hierarchicalTree.getLeafNodes())
				Collections.sort(node.leafDistribution.condSet, ec);
		}
		
		int width = 3 * x0 + this.hierarchicalTree.leafDistribution.condSet.size() 
			+ this.nonInformCond.leafDistribution.condSet.size();
		int height;
		int x = x0;
		int ybase = y0;

		if (this.hierarchicalTree.nodeStatus.equals("leaf")){ // no tree
			height = 2 * y0 + this.genes.size()*2;
		}	
		else {
			// coordinates of root
			x = x0 + this.nonInformCond.leafDistribution.condSet.size() + this.hierarchicalTree.leftChild.leafDistribution.condSet.size();
			ybase = y0 + (h + 4) * (this.hierarchicalTree.treeDepth() + 1);
			height = ybase + this.genes.size()*2 + y0;
		}
		BufferedImage image = new BufferedImage(sc * width, sc * height,
				BufferedImage.TYPE_INT_BGR);
		Graphics2D g2;
		g2 = image.createGraphics();
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		g2.scale(sc, sc);
		g2.setStroke(new BasicStroke(0.5F));
		if (this.hierarchicalTree.nodeStatus.equals("leaf")) { // no tree
			Collections.sort(this.hierarchicalTree.leafDistribution.condSet);
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			// draw data
			draw1leafExtra(geneSet, x0, y0, g2);
		} else {
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			drawLeaf(this.nonInformCond, x0, ybase, g2);
			drawnodeExtra(geneSet, this.hierarchicalTree, x, y0, h, ybase, g2);
		}
		File outFile = new File(filename);
		try {
			ImageIO.write(image, "png", outFile);
		} catch (IOException e) {
			return;
		}
	}

	/**
	 * Draws a figure of the current regulation program for this module, same as draw() method
	 * plus the conditions names.
	 * 
	 * The names are in the conditionSet ArrayList
	 * 
	 * @param filename name of file to store figure
	 * 
	 * @author tomic, stmae, erbon
	 * 
	 */
	public void drawConditions(String filename) {
		// scale factor
		// Increase the scale, otherwise it's impossible to read the conditions names
		int sc = 12;
		// coordinates of origin
		int x0 = 10;
		int y0 = 10;
		// tree step height
		int h = 18;
		// figure width
		int width = 3 * x0 + this.hierarchicalTree.leafDistribution.condSet.size();
		int height;
		int x = x0;
		int ybase = y0;

		if (this.hierarchicalTree.nodeStatus.equals("leaf")){ // no tree
			height = 2 * y0 + this.genes.size()*2;
		}	
		else {
			// coordinates of root
			x = x0 + this.hierarchicalTree.leftChild.leafDistribution.condSet.size();
			ybase = y0 + (h + 4) * (this.hierarchicalTree.treeDepth() + 1);
			height = ybase + this.genes.size()*2 + y0;
		}
		BufferedImage image = new BufferedImage(sc * width, sc * height, BufferedImage.TYPE_INT_BGR);
		Graphics2D g2;
		g2 = image.createGraphics();
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.scale(sc, sc);
		g2.setStroke(new BasicStroke(0.5F));
		if (this.hierarchicalTree.nodeStatus.equals("leaf")) { // no tree
			Collections.sort(this.hierarchicalTree.leafDistribution.condSet);
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			// draw data
			draw1leaf(x0, y0, g2);
		} else {
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			drawnode(this.hierarchicalTree, x, y0, h, ybase, g2);
		}

		// Draw the names for each condition, at the bottom of the image
		int x_exp;
		int y_exp;
		y_exp =  height - 8;
		g2.setFont(new Font("SansSerif", Font.PLAIN, 1));
		g2.setPaint(Color.black);
		for (int m = 0; m < this.hierarchicalTree.leafDistribution.condSet.size(); m++) {
			int cond = this.hierarchicalTree.leafDistribution.condSet.get(m);
			String condName = this.moduleNetwork.conditionSet.get(cond).name;
			// draw the string vertically
			// to draw it 45 deg, use PI/3
			g2.rotate(Math.PI/2,x0+m, y_exp);
			g2.drawString(condName, x0 + m, y_exp);
			g2.rotate(-Math.PI/2,x0+m, y_exp);
		}

		File outFile = new File(filename);
		try {
			ImageIO.write(image, "png", outFile);
		} catch (IOException e) {
			return;
		}
	}
	
	/**
	 * Converts an expression value to a color. For ratio data, color shades are blue for over-expression
	 * and yellow for under-expression with black to interpolate. For non-ratio Affymetrix data, color shades
	 * go from blue (lowest expression level) to yellow (highest expression level) (just to be logical :-).
	 * The amount of blue or yellow hue can depend on the particular gene whose expression is considered.
	 * See also {@link Gene#upProb(double, double, double)} and {@link Gene#downProb(double, double, double)}.
	 * 
	 * @param gene gene whose expression is considered.
	 * @param data expression value.
	 * 
	 * @author tomic
	 * 
	 */
	public Color data2Color(Gene gene, double data) {
		float downPerc=0.5F, upPerc=0.5F;
		Color col = Color.black;
		downPerc = new Float(gene.downProb(data,this.moduleNetwork.dataMean,this.moduleNetwork.dataSigma));
		upPerc = new Float(gene.upProb(data,this.moduleNetwork.dataMean,this.moduleNetwork.dataSigma));
		//col = new Color(upPerc, upPerc, downPerc);
		col = new Color(downPerc, upPerc, 0.0F);
//		switch (this.moduleNetwork.dataType){
//		case RATIO:
//			downPerc = new Float(gene.downProb(data,this.moduleNetwork.dataMean,this.moduleNetwork.dataSigma));
//			upPerc = new Float(gene.upProb(data,this.moduleNetwork.dataMean,this.moduleNetwork.dataSigma));
//			col = new Color(upPerc, upPerc, downPerc);
//		case AFFY:
//			double x;
//			if (data <= this.moduleNetwork.dataMean)
//				x = Math.max((data-this.moduleNetwork.dataMean)/(1.5*this.moduleNetwork.dataSigma), -0.5);
//			else
//				x = Math.min((data-this.moduleNetwork.dataMean)/(1.5*this.moduleNetwork.dataSigma), 0.5);
//			upPerc = new Float(0.5 + x);
//			downPerc = 1.0F - upPerc;
//			col = new Color(upPerc, upPerc, 0);
//		}
		return col;
	}

	/**
	 * Converts an expression value to a color. Intermediate values are a gradient between blue and yellow.
	 * Mean and sigma are passed are parameters
	 * 
	 * @param value
	 * @param mean
	 * @param sigma
	 * @return Color object
	 * 
	 * @author erbon
	 */
	public Color data2Color(double value, double mean, double sigma) {
		float yellowPerc=0.5F, bluePerc=0.5F;
		double x;

		if (value <= mean)
			x = Math.max((value - mean)/(1.5 * sigma), -0.5);
		else
			x = Math.min((value - mean)/(1.5 * sigma), 0.5);
		
		yellowPerc = new Float(0.5 + x);
		bluePerc = 1.0F - yellowPerc;	
	
		return new Color(yellowPerc, yellowPerc, bluePerc);
	}
	
	/**
	 * Draws a node in the hierarchical tree. Draws the regulator expression values
	 * at a node in the tree as well as the lines linking to other nodes. For a leaf
	 * node it draws the expression values for the experiments at the node.
	 * 
	 * @param node TreeNode to be drawn.
	 * @param x <em>x</em>-coordinate of start position for vertical node line.
	 * @param y <em>y</em>-coordinate of start position for vertical node line.
	 * @param h height of vertical node line
	 * @param ybase <em>y</em>-coordinate of baseline where leaves begin.
	 * @param g graphics in which node is drawn.
	 * 
	 * @author tomic
	 * 
	 */
	public void drawnode(TreeNode node, int x, int y, int h, int ybase,	Graphics2D g) {
		double[][] data = this.moduleNetwork.data; 
		if (node.nodeStatus.equals("internal")) {
			// where to start drawing regulator expression values
			int x1 = x - node.leftChild.leafDistribution.condSet.size();
			// draw regulator expression values left of split point
			for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
				Integer exp = node.leftChild.leafDistribution.condSet.get(m);
				if (!Double.isNaN(data[node.regulationSplit.regulator.number][exp])) {
					// convert data value to color
					Color col = data2Color(node.regulationSplit.regulator,node.regulationSplit.regulator.normData[exp]);
					g.setPaint(col);
				} else {
					// missing value
					g.setPaint(Color.white);
				}
				g.fillRect(x1 + m, y + h - 4, 1, 2);
			}
			// draw regulator expression values right of split point
			for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
				int exp = node.rightChild.leafDistribution.condSet.get(m);
				if (!Double.isNaN(data[node.regulationSplit.regulator.number][exp])) {
					// convert data value to color
					Color col = data2Color(node.regulationSplit.regulator,node.regulationSplit.regulator.normData[exp]);
					g.setPaint(col);
				} else
					// missing value
					g.setPaint(Color.white);
				g.fillRect(x + m, y + h - 4, 1, 2);
			}
			g.setPaint(Color.black);
			// draw vertical split line
			g.drawLine(x, y, x, y + h);
			// set font for writing regulator name
			g.setFont(new Font("SansSerif", Font.BOLD, 4));
			FontRenderContext frc = g.getFontRenderContext();
			// find dimensions of regulator name
			String[] reg = node.regulationSplit.regulator.name.split("_");
			LineMetrics metrics = g.getFont().getLineMetrics(reg[0], frc);
			float regwidth = (float) g.getFont().getStringBounds(reg[0], frc).getWidth();
			float regheight = (float) g.getFont().getStringBounds(reg[0], frc).getHeight();
//			LineMetrics metrics = g.getFont().getLineMetrics(node.regulationSplit.regulator.name, frc);
//			float regwidth = (float) g.getFont().getStringBounds(node.regulationSplit.regulator.name, frc).getWidth();
//			float regheight = (float) g.getFont().getStringBounds(node.regulationSplit.regulator.name, frc).getHeight();
			// draw ellipse around regulator name
			Shape r = new Ellipse2D.Double(x - 0.7 * regwidth, y + h - 11 - 1.4
					* regheight + 1.5 * metrics.getDescent(), 1.4 * regwidth,
					1.6 * regheight);
			if (this.genes.contains(node.regulationSplit.regulator))
				g.setPaint(Color.yellow);
			else
				g.setPaint(Color.white);
			g.fill(r);
			g.setPaint(Color.black);
			g.draw(r);
			// write regulator name
			g.drawString(reg[0], new Float(x - 0.5 * regwidth), new Float(y + h - 11));
//			g.drawString(node.regulationSplit.regulator.name, new Float(x - 0.5 * regwidth),
//					new Float(y + h - 11));
			// annotate split with its score gain and regulator assignment entropy
			double diff = (-node.leafDistribution.score
					+ node.leftChild.leafDistribution.score + node.rightChild.leafDistribution.score)
					/ this.moduleNetwork.numGenes;
			double regscore = Math.exp(node.regulationSplit.regulatorScore/(double)node.leafDistribution.condSet.size()); 
			//String quality = String.format("(%.2g,%.2g)", diff, regscore);
			String quality = String.format("%.2g", regscore);
			g.setFont(new Font("SansSerif", Font.BOLD, 3));
			//g.drawString(quality, new Float(x - 8), new Float(y + h + 3));
			g.drawString(quality, new Float(x-2), new Float(y + h + 3));
			// x-coordinate for leftChild
			int xl;
			if (node.leftChild.nodeStatus.equals("internal")) {
				// if child is also internal, start should be in the middle of its own children
				xl = x - node.leftChild.rightChild.leafDistribution.condSet.size();
			} else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				xl = x - node.leftChild.leafDistribution.condSet.size() / 2;
				g.drawLine(xl, y + h, xl, ybase);
				int xl1 = x - node.leftChild.leafDistribution.condSet.size();
				int gn = 0;
				// draw the data points
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							Color col = data2Color(gene, data[gene.number][exp]);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, ybase + 2*gn, 1, 2);
					}
					gn++;
				}
				g.setPaint(Color.black);
				// draw frame around expression values
				g.drawRect(xl1, ybase, node.leftChild.leafDistribution.condSet.size(), this.genes.size()*2);
				// print average posterior probability of leaf
			}
			// x-coordinate for rightChild
			int xr;
			if (node.rightChild.nodeStatus.equals("internal"))
				// if child is also internal, start should be in the middle of its own children
				xr = x + node.rightChild.leftChild.leafDistribution.condSet.size();
			else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				xr = x + node.rightChild.leafDistribution.condSet.size() / 2;
				g.drawLine(xr, y + h, xr, ybase);
				int gn = 0;
				// draw the data points
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							Color col = data2Color(gene, data[gene.number][exp]);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, ybase + 2*gn, 1, 2);
					}
					if(node == this.hierarchicalTree.rightMostNode()){
						//add names of genes to plot
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						g.drawString(gene.name.split("_")[0], new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(ybase + 2*gn+2));
					}	
					gn++;
				}
				g.setPaint(Color.black);
				// draw frame around expression values
				g.drawRect(x, ybase, node.rightChild.leafDistribution.condSet.size(), this.genes.size()*2);
			}
			// draw horizontal line between start coordinates for the children
			g.drawLine(xl, y + h, xr, y + h);
			// draw the children, right one is drawn lower then left one for clarity
			drawnode(node.leftChild, xl, y + h, h, ybase, g);
			drawnode(node.rightChild, xr, y + h, h + 4, ybase, g);
			// draw line between leaves
			g.setPaint(Color.yellow);
			//g.setPaint(new Color(1.0F, 0.0F, 0.5F));
			g.drawLine(x, ybase, x, ybase + this.genes.size()*2);
			g.setPaint(Color.black);
		}
	}
	
	/**
	 * Draws a node in the hierarchical tree with extra information. Draws the regulator expression values
	 * at a node in the tree as well as the lines linking to other nodes. For a leaf
	 * node it draws the expression values for the experiments at the node.
	 * 
	 * @param node TreeNode to be drawn.
	 * @param x <em>x</em>-coordinate of start position for vertical node line.
	 * @param y <em>y</em>-coordinate of start position for vertical node line.
	 * @param h height of vertical node line
	 * @param ybase <em>y</em>-coordinate of baseline where leaves begin.
	 * @param g graphics in which node is drawn.
	 * 
	 * @author tomic
	 */
	public void drawnodeExtra(ArrayList<Gene> geneSet, TreeNode node, int x, int y, int h, int ybase,
			Graphics2D g) {
		double[][] data = this.moduleNetwork.data; 
		if (node.nodeStatus.equals("internal")) {
			int xL = x - node.leftChild.leafDistribution.condSet.size();
			int xR = x + node.rightChild.leafDistribution.condSet.size();
			TreeNode par = node;
			while (par.parentNode instanceof TreeNode){
				if (par.equals(par.parentNode.rightChild))
					xL = xL - par.parentNode.leftChild.leafDistribution.condSet.size();
				else
					xR = xR + par.parentNode.rightChild.leafDistribution.condSet.size();
				par = par.parentNode;
			}
			for (int m=0; m<this.hierarchicalTree.leafDistribution.condSet.size(); m++){
				int exp = this.hierarchicalTree.leafDistribution.condSet.get(m);
				if (!Double.isNaN(data[node.regulationSplit.regulator.number][exp])) {
					// convert data value to color
					Color col = data2Color(node.regulationSplit.regulator,node.regulationSplit.regulator.normData[exp]);
					g.setPaint(col);
				} else {
					// missing value
					g.setPaint(Color.white);
				}
				g.fillRect(xL + m, y + h - 4, 1, 2);
			}
			for (int m=0; m<this.nonInformCond.leafDistribution.condSet.size(); m++){
				int exp = this.nonInformCond.leafDistribution.condSet.get(m);
				if (!Double.isNaN(data[node.regulationSplit.regulator.number][exp])) {
					// convert data value to color
					Color col = data2Color(node.regulationSplit.regulator,node.regulationSplit.regulator.normData[exp]);
					g.setPaint(col);
				} else {
					// missing value
					g.setPaint(Color.white);
				}
				g.fillRect(xL - this.nonInformCond.leafDistribution.condSet.size() + m, y + h - 4, 1, 2);
			}
			g.setPaint(Color.red);
			g.drawRect(x-node.leftChild.leafDistribution.condSet.size(), y+h-4, node.leafDistribution.condSet.size(), 2);
			g.setPaint(Color.black);
			// draw vertical split line
			g.drawLine(x, y, x, y + h);
			// set font for writing regulator name
			g.setFont(new Font("SansSerif", Font.BOLD, 4));
			FontRenderContext frc = g.getFontRenderContext();
			// find dimensions of regulator name
			LineMetrics metrics = g.getFont().getLineMetrics(node.regulationSplit.regulator.name, frc);
			float regwidth = (float) g.getFont().getStringBounds(node.regulationSplit.regulator.name, frc).getWidth();
			float regheight = (float) g.getFont().getStringBounds(node.regulationSplit.regulator.name, frc).getHeight();
			// draw ellipse around regulator name
			Shape r = new Ellipse2D.Double(x - 0.7 * regwidth, y + h - 11 - 1.4
					* regheight + 1.5 * metrics.getDescent(), 1.4 * regwidth,
					1.6 * regheight);
			if (this.genes.contains(node.regulationSplit.regulator))
				g.setPaint(Color.yellow);
			else
				g.setPaint(Color.white);
			g.fill(r);
			g.setPaint(Color.black);
			g.draw(r);
			// write regulator name
			g.drawString(node.regulationSplit.regulator.name, new Float(x - 0.5 * regwidth),
					new Float(y + h - 11));
			// annotate split with its score gain and regulator assignment entropy
			double diff = (-node.leafDistribution.score
					+ node.leftChild.leafDistribution.score + node.rightChild.leafDistribution.score)
					/ this.moduleNetwork.numGenes;
			double regscore = Math.exp(node.regulationSplit.regulatorScore/(double)node.leafDistribution.condSet.size()); 
			String quality = String.format("(%.2g,%.2g)", diff, regscore);
			g.setFont(new Font("SansSerif", Font.BOLD, 3));
			g.drawString(quality, new Float(x - 8), new Float(y + h + 3));
			// x-coordinate for leftChild
			int xl;
			if (node.leftChild.nodeStatus.equals("internal")) {
				// if child is also internal, start should be in the middle of
				// its own children
				xl = x - node.leftChild.rightChild.leafDistribution.condSet.size();
			} else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				xl = x - node.leftChild.leafDistribution.condSet.size() / 2;
				g.drawLine(xl, y + h, xl, ybase);
				int xl1 = x - node.leftChild.leafDistribution.condSet.size();
				int gn = 0;
				// draw the data points
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							Color col = data2Color(gene, data[gene.number][exp]);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, ybase + 2*gn, 1, 2);
					}
					gn++;
				}
				g.setPaint(Color.red);
				// draw frame around expression values
				g.drawRect(xl1, ybase, node.leftChild.leafDistribution.condSet.size(), this.genes.size()*2);
				// print 
				g.setPaint(Color.black);
				g.setFont(new Font("SansSerif", Font.BOLD, 4));
				double p = node.leftChild.leafDistribution.gaussianParam[1];
				String scr =  String.format("%.2f", p);
				metrics = g.getFont().getLineMetrics(scr, frc);
				float scrw = (float) g.getFont().getStringBounds(scr, frc).getWidth();
				g.drawString(scr, new Float(xl-0.5*scrw), new Float(ybase+2*this.genes.size()+5));
				
				scr =  String.format("%.2f", node.leftChild.leafDistribution.gaussianParam[0]);
				metrics = g.getFont().getLineMetrics(scr, frc);
				scrw = (float) g.getFont().getStringBounds(scr, frc).getWidth();
				g.drawString(scr, new Float(xl-0.5*scrw), new Float(ybase+2*this.genes.size()+9.5));
			}
			// x-coordinate for rightChild
			int xr;
			if (node.rightChild.nodeStatus.equals("internal"))
				// if child is also internal, start should be in the middle of
				// its own children
				xr = x + node.rightChild.leftChild.leafDistribution.condSet.size();
			else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				xr = x + node.rightChild.leafDistribution.condSet.size() / 2;
				g.drawLine(xr, y + h, xr, ybase);
				int gn = 0;
				// draw the data points
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							Color col = data2Color(gene, data[gene.number][exp]);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, ybase + 2*gn, 1, 2);
					}
					if(node == this.hierarchicalTree.rightMostNode()){
						//add names of genes to plot, highlight if in specific set
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						if (geneSet.contains(gene))
							g.setPaint(Color.red);
						else
							g.setPaint(Color.black);
						g.drawString(gene.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(ybase + 2*gn+2));
					}	
					gn++;
				}
				g.setPaint(Color.black);
				// draw frame around expression values
				g.setPaint(Color.red);
				g.drawRect(x, ybase, node.rightChild.leafDistribution.condSet.size(), this.genes.size()*2);
				// print 
				g.setPaint(Color.black);
				g.setFont(new Font("SansSerif", Font.BOLD, 4));
				double p = node.rightChild.leafDistribution.gaussianParam[1];
				String scr =  String.format("%.2f", p);
				metrics = g.getFont().getLineMetrics(scr, frc);
				float scrw = (float) g.getFont().getStringBounds(scr, frc).getWidth();
				g.drawString(scr, new Float(xr-0.5*scrw), new Float(ybase+2*this.genes.size()+5));
				
				scr =  String.format("%.2f", node.rightChild.leafDistribution.gaussianParam[0]);
				metrics = g.getFont().getLineMetrics(scr, frc);
				scrw = (float) g.getFont().getStringBounds(scr, frc).getWidth();
				g.drawString(scr, new Float(xr-0.5*scrw), new Float(ybase+2*this.genes.size()+9.5));
			}
			// draw horizontal line between start coordinates for the children
			g.drawLine(xl, y + h, xr, y + h);
			// draw the children, right one is drawn lower then left one for clarity
			drawnodeExtra(geneSet, node.leftChild, xl, y + h, h, ybase, g);
			drawnodeExtra(geneSet, node.rightChild, xr, y + h, h + 4, ybase, g);
			// draw line between leaves
			//g.setPaint(Color.red);
			g.setPaint(new Color(1.0F, 0.0F, 0.5F));
			g.drawLine(x, ybase, x, ybase + this.genes.size()*2);
			g.setPaint(Color.black);
		}
	}


	/**
	 * Draw module figure as a png, with top regulators, GO categories and microRNA targets.
	 * 
	 * @param filename Ouput file name.
	 * 
	 * @author erbon
	 */
	public void drawMir(String filename) {

		int sc = 15;        //scale factor, make it big for very small fonts or increased graphic quality
		int x0 = 10;		// coordinates of origin
		int y0 = 10;
		int h = 1;			// tree step height
		int width;			// figure width
		int height;			// figure height
		int x = x0;
		int ybase = y0;

		// get max text size for condition names and GO categories
		// create dummy image to estimate text size
		BufferedImage tmp_img = new BufferedImage(200,200, BufferedImage.TYPE_INT_BGR);
		Graphics2D tmp_g;
		tmp_g = tmp_img.createGraphics();
		tmp_g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		tmp_g.scale(sc, sc);
		FontRenderContext frc = tmp_g.getFontRenderContext();
		float MaxTextLen = 0;
		float CondTextLen = 0;
		// get max condition name length
		for (int i=0;i<this.hierarchicalTree.leafDistribution.condSet.size();i++) {
			int cond_num = this.hierarchicalTree.leafDistribution.condSet.get(i);
			String cond_name = this.moduleNetwork.conditionSet.get(cond_num).name;
			CondTextLen = (float) (new Font("SansSerif",Font.PLAIN,1)).getStringBounds(cond_name, frc).getWidth();
			if (CondTextLen > MaxTextLen)
				MaxTextLen = CondTextLen;
		}
		// get max GO category name length
		for (OntologyTerm o : this.GOsuper) {
			String cat_name = o.getName();
			CondTextLen = (float) (new Font("SansSerif",Font.PLAIN,1)).getStringBounds(cat_name, frc).getWidth();
			if (CondTextLen > MaxTextLen)
				MaxTextLen = CondTextLen;
		}
		
		// set width according to the number of conditions, top regulators and GO super categories
		width = 3 * x0 + this.hierarchicalTree.leafDistribution.condSet.size() + this.topRegulators.size() + this.GOsuper.size() + 10;

		// set height if there is no reg. tree / there is a reg. tree
		if (this.hierarchicalTree.nodeStatus.equals("leaf")){
			height = 2 * y0 + this.genes.size()*2;
		}
		else {
			x = x0 + this.hierarchicalTree.leftChild.leafDistribution.condSet.size();
			//ybase = y0 + (h + 4) * (this.hierarchicalTree.treeDepth() + 1);
			ybase = y0 + h * (this.hierarchicalTree.treeDepth() + 1) + this.topRegulators.size()*2 + 2;
			//height = ybase + this.genes.size()*2 + this.topRegulators.size()*2 + y0;
			height = ybase + this.genes.size()*2 + y0 + (int) MaxTextLen;
		}

		// create graphics2D object
		BufferedImage image = new BufferedImage(sc * width, sc * height, BufferedImage.TYPE_INT_BGR);
		Graphics2D g2;
		g2 = image.createGraphics();
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.scale(sc, sc);
		//g2.setStroke(new BasicStroke(0.2F));

		// if there is no reg. tree, draw one big leaf, else draw node by node
		if (this.hierarchicalTree.nodeStatus.equals("leaf")) {
//			Collections.sort(this.hierarchicalTree.leafDistribution.condSet);
//			g2.setPaint(Color.white);
//			g2.fillRect(0, 0, width, height);
//			// draw data
//			draw1leaf(x0, y0, g2);	
		} else {
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			drawNodeMir(this.hierarchicalTree, x, y0, h, ybase, g2);
		}

		File outFile = new File(filename);
		try {
			ImageIO.write(image, "png", outFile);
		} catch (IOException e) {
			return;
		}
	}

	/**
	 * Draw module figure as eps file, with top regulators, GO categories and microRNA targets.
	 * 
	 * @param filename Output file name.
	 * 
	 * @author erbon
	 */
	public void drawMirEps(String filename) {

		int sc = 4;         //scale factor
		int x0 = 10;		// coordinates of origin
		int y0 = 10;
		int h = 1;			// tree step height
		int width;			// figure width
		int height;			// figure height
		int x = x0;
		int ybase = y0;

		// get max text size for condition names and GO categories
		// create dummy image to estimate text size
		BufferedImage tmp_img = new BufferedImage(200,200, BufferedImage.TYPE_INT_BGR);
		Graphics2D tmp_g;
		tmp_g = tmp_img.createGraphics();
		tmp_g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		tmp_g.scale(sc, sc);
		FontRenderContext frc = tmp_g.getFontRenderContext();
		float MaxTextLen = 0;
		float CondTextLen = 0;
		// get max condition name length
		for (int i=0;i<this.hierarchicalTree.leafDistribution.condSet.size();i++) {
			int cond_num = this.hierarchicalTree.leafDistribution.condSet.get(i);
			String cond_name = this.moduleNetwork.conditionSet.get(cond_num).name;
			CondTextLen = (float) (new Font("SansSerif",Font.PLAIN,1)).getStringBounds(cond_name, frc).getWidth();
			if (CondTextLen > MaxTextLen)
				MaxTextLen = CondTextLen;
		}
		// get max GO category name length
		for (OntologyTerm o : this.GOsuper) {
			String cat_name = o.getName();
			CondTextLen = (float) (new Font("SansSerif",Font.PLAIN,1)).getStringBounds(cat_name, frc).getWidth();
			if (CondTextLen > MaxTextLen)
				MaxTextLen = CondTextLen;
		}
		
		// set width according to the number of conditions, top regulators and GO super categories
		width = 3 * x0 + this.hierarchicalTree.leafDistribution.condSet.size() + this.topRegulators.size() + this.GOsuper.size() + 10;

		// set height if there is no reg. tree / there is a reg. tree
		if (this.hierarchicalTree.nodeStatus.equals("leaf")){
			height = 2 * y0 + this.genes.size()*2;
		}
		else {
			x = x0 + this.hierarchicalTree.leftChild.leafDistribution.condSet.size();
			//ybase = y0 + (h + 4) * (this.hierarchicalTree.treeDepth() + 1);
			ybase = y0 + h * (this.hierarchicalTree.treeDepth() + 1) + this.topRegulators.size()*2 + 2;
			//height = ybase + this.genes.size()*2 + this.topRegulators.size()*2 + y0;
			height = ybase + this.genes.size()*2 + y0 + (int) MaxTextLen;
		}

		// create graphics2D object
//		BufferedImage image = new BufferedImage(sc * width, sc * height, BufferedImage.TYPE_INT_BGR);
//		Graphics2D g2;
//		g2 = image.createGraphics();
//		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
//		g2.scale(sc, sc);

		try{
			EpsGraphics g2 = new EpsGraphics(this.name, 
					new FileOutputStream(new File(filename + ".eps")),
					0,
					0,
					sc*width,
					sc*height,
					ColorMode.COLOR_CMYK);
			g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			g2.scale(sc, sc);

			// if there is no reg. tree, draw one big leaf, else draw node by node
			if (this.hierarchicalTree.nodeStatus.equals("leaf")) {
//				Collections.sort(this.hierarchicalTree.leafDistribution.condSet);
//				g2.setPaint(Color.white);
//				g2.fillRect(0, 0, width, height);
//				// draw data
//				draw1leaf(x0, y0, g2);	
			} else {
				g2.setPaint(Color.white);
				g2.fillRect(0, 0, width, height);
				drawNodeMir(this.hierarchicalTree, x, y0, h, ybase, g2);
				
				g2.setPaint(Color.black);
				g2.setFont(new Font("SansSerif", Font.PLAIN, 1));
				g2.drawString(filename, 10, 3);
				
			}
			
			g2.flush();
			g2.close();
		}
		catch (Exception e){
			e.printStackTrace();
			System.out.println(e);
			System.exit(1);
		}
	}

	/**
	 * Draw module figure as eps file, with top regulators and GO categories
	 * 
	 * @param filename Output file name.
	 * 
	 * @author erbon
	 */
	public void drawGOEps(String filename) {

		int sc = 4;         //scale factor
		int x0 = 10;		// coordinates of origin
		int y0 = 10;
		int h = 1;			// tree step height
		int width;			// figure width
		int height;			// figure height
		int x = x0;
		int ybase = y0;

		// get max text size for condition names and GO categories
		// create dummy image to estimate text size
		BufferedImage tmp_img = new BufferedImage(200,200, BufferedImage.TYPE_INT_BGR);
		Graphics2D tmp_g;
		tmp_g = tmp_img.createGraphics();
		tmp_g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		tmp_g.scale(sc, sc);
		FontRenderContext frc = tmp_g.getFontRenderContext();
		float MaxTextLen = 0;
		float CondTextLen = 0;
		// get max condition name length
		for (int i=0;i<this.hierarchicalTree.leafDistribution.condSet.size();i++) {
			int cond_num = this.hierarchicalTree.leafDistribution.condSet.get(i);
			String cond_name = this.moduleNetwork.conditionSet.get(cond_num).name;
			CondTextLen = (float) (new Font("SansSerif",Font.PLAIN,1)).getStringBounds(cond_name, frc).getWidth();
			if (CondTextLen > MaxTextLen)
				MaxTextLen = CondTextLen;
		}
		// get max GO category name length
		for (OntologyTerm o : this.GOsuper) {
			String cat_name = o.getName();
			CondTextLen = (float) (new Font("SansSerif",Font.PLAIN,1)).getStringBounds(cat_name, frc).getWidth();
			if (CondTextLen > MaxTextLen)
				MaxTextLen = CondTextLen;
		}
		
		// set width according to the number of conditions, top regulators and GO super categories
		width = 3 * x0 + this.hierarchicalTree.leafDistribution.condSet.size() + this.topRegulators.size() + this.GOsuper.size() + 10;

		// set height if there is no reg. tree / there is a reg. tree
		if (this.hierarchicalTree.nodeStatus.equals("leaf")){
			height = 2 * y0 + this.genes.size()*2;
		}
		else {
			x = x0 + this.hierarchicalTree.leftChild.leafDistribution.condSet.size();
			//ybase = y0 + (h + 4) * (this.hierarchicalTree.treeDepth() + 1);
			ybase = y0 + h * (this.hierarchicalTree.treeDepth() + 1) + this.topRegulators.size()*2 + 2;
			//height = ybase + this.genes.size()*2 + this.topRegulators.size()*2 + y0;
			height = ybase + this.genes.size()*2 + y0 + (int) MaxTextLen;
		}

		try{
			EpsGraphics g2 = new EpsGraphics(this.name, 
					new FileOutputStream(new File(filename + ".eps")),
					0,
					0,
					sc*width,
					sc*height,
					ColorMode.COLOR_CMYK);
			g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			g2.scale(sc, sc);

			if (this.hierarchicalTree.nodeStatus.equals("leaf")) {
				// don't do anything
			} else {
				g2.setPaint(Color.white);
				g2.fillRect(0, 0, width, height);
				drawNodeGO(this.hierarchicalTree, x, y0, h, ybase, g2);
				
				g2.setPaint(Color.black);
				g2.setFont(new Font("SansSerif", Font.PLAIN, 1));
				g2.drawString(filename, 10, 3);
				
			}
			
			g2.flush();
			g2.close();
		}
		catch (Exception e){
			e.printStackTrace();
			System.out.println(e);
			System.exit(1);
		}
	}


	/**
	 * Draw module figure as eps file, with top regulators and one tree, Eps format.
	 * 
	 * @param filename Output file name.
	 * 
	 * @author erbon
	 */
	public void drawEps(String filename) {

		int sc = 4;         //scale factor
		int x0 = 10;		// coordinates of origin
		int y0 = 10;
		int h = 1;			// tree step height
		int width;			// figure width
		int height;			// figure height
		int x = x0;
		int ybase = y0;

		// get max text size for condition names
		// create dummy image to estimate text size
		BufferedImage tmp_img = new BufferedImage(200,200, BufferedImage.TYPE_INT_BGR);
		Graphics2D tmp_g;
		tmp_g = tmp_img.createGraphics();
		tmp_g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		tmp_g.scale(sc, sc);
		FontRenderContext frc = tmp_g.getFontRenderContext();
		float MaxTextLen = 1;
		float CondTextLen = 1;
		// get max condition name length
		for (int i=0;i<this.hierarchicalTree.leafDistribution.condSet.size();i++) {
			int cond_num = this.hierarchicalTree.leafDistribution.condSet.get(i);
			String cond_name = this.moduleNetwork.conditionSet.get(cond_num).name;
			CondTextLen = (float) (new Font("SansSerif",Font.PLAIN,1)).getStringBounds(cond_name, frc).getWidth();
			if (CondTextLen > MaxTextLen)
				MaxTextLen = CondTextLen;
		}
		
		float MaxGeneNameLen = 1;
		for (int i=0;i < this.genes.size(); i++) {
			Gene g = this.genes.get(i);
			float len = (float) (new Font("SansSerif",Font.BOLD,2)).getStringBounds(g.name, frc).getWidth();
			if (len > MaxGeneNameLen)
				MaxGeneNameLen = len;
		}
		
		// set width according to the number of conditions, top regulators
		width = 2 * x0 + this.hierarchicalTree.leafDistribution.condSet.size() + this.topRegulators.size() + 5 + (int) MaxGeneNameLen;

		// set height if there is no reg. tree / there is a reg. tree
		if (this.hierarchicalTree.nodeStatus.equals("leaf")){
			height = y0 + this.genes.size() * 2 + 15 + (int) MaxTextLen;
		}
		else {
			x = x0 + this.hierarchicalTree.leftChild.leafDistribution.condSet.size();
			//ybase = y0 + (h + 4) * (this.hierarchicalTree.treeDepth() + 1);
			ybase = y0 + h * (this.hierarchicalTree.treeDepth() + 1) + this.topRegulators.size()*2 + 2;
			//height = ybase + this.genes.size()*2 + this.topRegulators.size()*2 + y0;
			height = ybase + this.genes.size()*2 + y0 + (int) MaxTextLen;
		}

		try{
			EpsGraphics g2 = new EpsGraphics(this.name, 
					new FileOutputStream(new File(filename + ".eps")),
					0,
					0,
					sc*width,
					sc*height,
					ColorMode.COLOR_CMYK);
			g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			g2.scale(sc, sc);

			//if (this.hierarchicalTree.nodeStatus.equals("leaf")) {
			if (this.topRegulators.size() == 0) {
				Collections.sort(this.hierarchicalTree.leafDistribution.condSet);
				g2.setPaint(Color.white);
				g2.fillRect(0, 0, width, height);
				// draw data
				drawOneLeafEps(x0, y0, g2);
			} else {
				g2.setPaint(Color.white);
				g2.fillRect(0, 0, width, height);
				drawNodeEps(this.hierarchicalTree, x, y0, h, ybase, g2);
			}
			
			// draw the module names in the upper right corner
			g2.setPaint(Color.black);
			g2.setFont(new Font("SansSerif", Font.PLAIN, 1));
			g2.drawString(filename, 10, 3);
			
			g2.flush();
			g2.close();
		}
		catch (Exception e){
			e.printStackTrace();
			System.out.println(e);
			System.exit(1);
		}
	}
	
	/**
	 * Draws regulatory tree leaves with top regulators, GO categories and microRNA targets.
	 * 
	 * @param node TreeNode to be drawn.
	 * @param x <em>x</em>-coordinate of start position for vertical node line.
	 * @param y <em>y</em>-coordinate of start position for vertical node line.
	 * @param h height of vertical node line
	 * @param ybase <em>y</em>-coordinate of baseline where leaves begin.
	 * @param g graphics in which node is drawn.
	 * 
	 * @author erbon
	 */
	public void drawNodeMir(TreeNode node, int x, int y, int h, int ybase,	Graphics2D g) {
		double[][] data = this.moduleNetwork.data; 
		if (node.nodeStatus.equals("internal")) {
			// where to start drawing regulator expression values
			int x1 = x - node.leftChild.leafDistribution.condSet.size();
			Stroke extraStr = new BasicStroke(1.0F);
//			// regulation tree ************************************************
//			// draw regulator expression values left of split point
//			for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
//			Integer exp = node.leftChild.leafDistribution.condSet.get(m);
//			if (!Double.isNaN(data[node.regulationSplit.regulator.number][exp])) {
//			// convert data value to color
//			Color col = data2Color(node.regulationSplit.regulator,data[node.regulationSplit.regulator.number][exp]);
//			g.setPaint(col);
//			} else {
//			// missing value
//			g.setPaint(Color.white);
//			}
//			g.fillRect(x1 + m, y + h - 4, 1, 2);
//			}
//			// draw regulator expression values right of split point
//			for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
//			int exp = node.rightChild.leafDistribution.condSet.get(m);
//			if (!Double.isNaN(data[node.regulationSplit.regulator.number][exp])) {
//			// convert data value to color
//			Color col = data2Color(node.regulationSplit.regulator,data[node.regulationSplit.regulator.number][exp]);
//			g.setPaint(col);
//			} else
//			// missing value
//			g.setPaint(Color.white);
//			g.fillRect(x + m, y + h - 4, 1, 2);
//			}

//			// draw vertical split line
			g.setPaint(Color.black);
			g.setStroke(new BasicStroke(1));
			g.drawLine(x, y, x, y + h);
//			// set font for writing regulator name
//			g.setFont(new Font("SansSerif", Font.BOLD, 4));
//			FontRenderContext frc = g.getFontRenderContext();
//			// find dimensions of regulator name
//			LineMetrics metrics = g.getFont().getLineMetrics(node.regulationSplit.regulator.name, frc);
//			float regwidth = (float) g.getFont().getStringBounds(node.regulationSplit.regulator.name, frc).getWidth();
//			float regheight = (float) g.getFont().getStringBounds(node.regulationSplit.regulator.name, frc).getHeight();
//			// draw ellipse around regulator name
//			Shape r = new Ellipse2D.Double(x - 0.7 * regwidth, y + h - 11 - 1.4
//			* regheight + 1.5 * metrics.getDescent(), 1.4 * regwidth,
//			1.6 * regheight);
//			if (this.genes.contains(node.regulationSplit.regulator))
//			g.setPaint(Color.yellow);
//			else
//			g.setPaint(Color.white);
//			g.fill(r);
//			g.setPaint(Color.black);
//			g.draw(r);
//			// write regulator name
//			g.drawString(node.regulationSplit.regulator.name, new Float(x - 0.5 * regwidth),
//			new Float(y + h - 11));
//			// annotate split with its score gain and regulator assignment entropy
//			double diff = (-node.leafDistribution.score
//			+ node.leftChild.leafDistribution.score + node.rightChild.leafDistribution.score)
//			/ this.moduleNetwork.numGenes;
//			double regscore = Math.exp(node.regulationSplit.regulatorScore/(double)node.leafDistribution.condSet.size()); 
//			String quality = String.format("(%.2g,%.2g)", diff, regscore);
//			g.setFont(new Font("SansSerif", Font.BOLD, 3));
//			g.drawString(quality, new Float(x - 8), new Float(y + h + 3));
//			// end regulation tree ********************************************


			// define y base for top regulators and condition labels
			int reg_ybase = (ybase - this.topRegulators.size() * 2 - 2);
			int text_ybase = ybase + this.genes.size() * 2 + 4;
			
			// leftChild ******************************************************
			int xl;
			if (node.leftChild.nodeStatus.equals("internal")) {
				// if child is also internal, start should be in the middle of its own children
				xl = x - node.leftChild.rightChild.leafDistribution.condSet.size();
			} else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				// = vertical line of the reg tree
				xl = x - node.leftChild.leafDistribution.condSet.size() / 2;
				g.setPaint(Color.black);
				g.setStroke(extraStr);
				g.drawLine(xl, y + h, xl, reg_ybase);
				int xl1 = x - node.leftChild.leafDistribution.condSet.size();
				int gn = 0;
				// draw the data points
				Collections.sort(node.leftChild.leafDistribution.condSet);
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							//Color col = data2Color(gene, data[gene.number][exp]);
							Color col = data2Color(data[gene.number][exp], this.mean, this.sigma);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, ybase + 2*gn, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(xl1 + m, ybase + 2*gn, 1, 2);
					}
					gn++;
				}
				
				// draw top regulators
				int ct=0;
				for (Gene reg : this.topRegulators) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[reg.number][exp])) {
							// convert data to color
							//Color col = data2Color(reg, data[reg.number][exp]);
							Color col = data2Color(data[reg.number][exp], reg.mean, reg.sigma);
							//******************************************************
							//   Option: plot value corresponding to how frequently 
							//   the regulator is assigned to a condition
							//******************************************************
							//Color col = data2Color(this.regConditions.get(reg).get(exp), 0.5, 0.1);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, reg_ybase + 2*ct, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(xl1 + m, reg_ybase + 2*ct, 1, 2);
					}
					ct++;
				}
				
				// draw black frame around module genes expression values
				//g.setPaint(Color.black);
				//g.drawRect(xl1, ybase, node.leftChild.leafDistribution.condSet.size(), this.genes.size()*2);
				//g.drawRect(xl1, reg_ybase, node.leftChild.leafDistribution.condSet.size(), this.topRegulators.size()*2);
				
				// draw condition labels
				for (int c=0;c<node.leftChild.leafDistribution.condSet.size();c++) {
					int num = node.leftChild.leafDistribution.condSet.get(c);
					String name = this.moduleNetwork.conditionSet.get(num).name;
					Color col = this.moduleNetwork.conditionSet.get(num).col;
					// draw rectangle for tissue type
					g.setPaint(col);
					g.fillRect(xl1 + c, text_ybase - 3, 1, 1);
					g.drawRect(xl1 + c, text_ybase - 3, 1, 1);
					if (this.moduleNetwork.conditionSet.get(num).state.equals("normal")) {
						g.setPaint(Color.lightGray);
						g.fillRect(xl1 + c, text_ybase - 2, 1, 1);
						g.drawRect(xl1 + c, text_ybase - 2, 1, 1);
					}
					else {
						g.setPaint(Color.darkGray);
						g.fillRect(xl1 + c, text_ybase - 2, 1, 1);
						g.drawRect(xl1 + c, text_ybase - 2, 1, 1);
					}
					//g.setPaint(Color.black);
					//g.setStroke(new BasicStroke(0.1F));
					//g.drawRect(xl1 + c, text_ybase - 3, 1, 1);
					//g.drawRect(xl1 + c, text_ybase - 2, 1, 1);
					// rotate canvas and draw string 45 deg.
					g.setPaint(Color.black);
					g.setFont(new Font("SansSerif", Font.PLAIN, 1));
					g.rotate(Math.PI/3, xl1 + c, text_ybase);
					g.drawString(name, xl1 + c, text_ybase);
					g.rotate(-Math.PI/3, xl1 + c, text_ybase);
				}
			}
			// end leftchild **************************************************

			// rightChild *****************************************************
			int xr;
			float MaxGeneNameLen = 0;
			float GeneNameLen=0;
			g.setFont(new Font("SansSerif", Font.BOLD, 2));
			FontRenderContext frc = g.getFontRenderContext();
			
			if (node.rightChild.nodeStatus.equals("internal"))
				// if child is also internal, start should be in the middle of its own children
				xr = x + node.rightChild.leftChild.leafDistribution.condSet.size();
			else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				// = vertical line of the reg tree
				xr = x + node.rightChild.leafDistribution.condSet.size() / 2;
				g.setPaint(Color.black);
				g.setStroke(extraStr);
				g.drawLine(xr, y + h, xr, reg_ybase);
				int gn = 0;
				// draw the data points
				Collections.sort(node.rightChild.leafDistribution.condSet);
				for (Gene gene : this.genes) {
					GeneNameLen =  (float) g.getFont().getStringBounds(gene.name, frc).getWidth();
					if (GeneNameLen > MaxGeneNameLen)
						MaxGeneNameLen = GeneNameLen;
					for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							//Color col = data2Color(gene, data[gene.number][exp]);
							Color col = data2Color(data[gene.number][exp], this.mean, this.sigma);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, ybase + 2*gn, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(x+m, ybase + 2*gn, 1, 2);
					}
//					if(node == this.hierarchicalTree.rightMostNode()){
//						//add names of genes to plot
//						g.setPaint(Color.black);
//						g.setFont(new Font("SansSerif", Font.BOLD, 2));
//						g.drawString(gene.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(ybase + 2*gn+2));
//					}	
					gn++;
				}
				
				// draw top regulators expression and names
				int ct=0;
				for (Gene reg : this.topRegulators) {
					GeneNameLen =  (float) g.getFont().getStringBounds(reg.name, frc).getWidth();
					if (GeneNameLen > MaxGeneNameLen)
						MaxGeneNameLen = GeneNameLen;
					for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[reg.number][exp])) {
							// convert data to color
							//Color col = data2Color(reg, data[reg.number][exp]);
							Color col = data2Color(data[reg.number][exp], reg.mean, reg.sigma);
							//*****************************************************
							//   Option: plot value corresponding to how frequently 
							//   the regulator is assigned to a condition
							//*****************************************************
							//Color col = data2Color(this.regConditions.get(reg).get(exp), 0.5, 0.1);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, reg_ybase + 2*ct, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(x+m, reg_ybase + 2*ct, 1, 2);
					}
//					if(node == this.hierarchicalTree.rightMostNode()){
//						//add names of genes to plot
//						g.setPaint(Color.black);
//						g.setFont(new Font("SansSerif", Font.BOLD, 2));
//						g.drawString(reg.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(reg_ybase + 2*ct+2));
//					}	
					ct++;
				}
				
				// draw extra information, like GO categories, microRNA seed matches
				if (node == this.hierarchicalTree.rightMostNode()) {
					
					// build the list of microRNA top regulators for this module
					ArrayList<Gene> MirRegList = new ArrayList<Gene>();
					for (Gene reg : this.topRegulators) {
						if (reg.name.startsWith("hsa"))
							MirRegList.add(reg);
					}
					
					// define extra information x coord. base
					int xtra_xbase = x + node.rightChild.leafDistribution.condSet.size() + (int) MaxGeneNameLen + 5;
					ct=0;
					for (Gene reg : this.topRegulators) {
						
						// draw top reg names
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						g.drawString(reg.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(reg_ybase + 2*ct+2));
						
						// draw microRNA seed target matches
						int ctm=0;
						for (Gene m : MirRegList) {
							
							boolean no_match = true;
							
							if (m.seedTargets.contains(reg)) {
								g.setPaint(new Color(87,191,227));  // light cyan
								no_match = false;
							} 
							
							if (m.predTargets.contains(reg)){
								g.setPaint(new Color(0,102,158)); //dark cyan
								no_match = false;
							}
							
							if (no_match)
								g.setPaint(Color.white);
							
							g.fillRect(xtra_xbase + this.GOsuper.size() + 2 + ctm, reg_ybase + 2*ct,1,2);
							g.setPaint(Color.black);
							g.setStroke(new BasicStroke(0.1F));
							g.drawRect(xtra_xbase + this.GOsuper.size() + 2 + ctm, reg_ybase + 2*ct,1,2);
							
							ctm++;
						}
						ct++;
					}
					
					ct=0;
					for (Gene gene : this.genes) {
						// draw gene name
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						g.drawString(gene.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(ybase + 2*ct+2));
						
						//draw GO categories
						int ctg = 0;
						for (OntologyTerm o : this.GOsuper) {
							if (gene.GOcat.contains(o)) {
								g.setPaint(Color.orange);
								g.fillRect(xtra_xbase + ctg, ybase + 2*ct,1,2);
								g.setPaint(Color.black);
								g.setStroke(new BasicStroke(0.1F));
								g.drawRect(xtra_xbase + ctg , ybase + 2*ct,1,2);
							}
							else {
								g.setPaint(Color.white);
								g.fillRect(xtra_xbase + ctg, ybase + 2*ct,1,2);
								g.setPaint(Color.black);
								g.setStroke(new BasicStroke(0.1F));
								g.drawRect(xtra_xbase + ctg, ybase + 2*ct,1,2);
							}
							ctg++;
						}
						
						// draw microRNA seed target matches
						int ctm=0;
						for (Gene m : MirRegList) {
							boolean no_match = true;
							if (m.seedTargets.contains(gene)) {
								g.setPaint(new Color(87,191,227)); // light cyan
								no_match = false;
							}
							if (m.predTargets.contains(gene)){
								g.setPaint(new Color(0,102,158)); //dark cyan
								no_match = false;
							}
							if (no_match) {
								g.setPaint(Color.white);
							}
							
							g.fillRect(xtra_xbase + this.GOsuper.size() + 2 + ctm, ybase + 2*ct,1,2);
							g.setPaint(Color.black);
							g.setStroke(new BasicStroke(0.1F));
							g.drawRect(xtra_xbase + this.GOsuper.size() + 2 + ctm, ybase + 2*ct,1,2);
							
							ctm++;
						}
						ct++;
					}
					
					// draw GO labels
					ct=0;
					for (OntologyTerm o : this.GOsuper) {
						// rotate canvas and draw string 45/90 deg.
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.PLAIN, 1));
						g.rotate(Math.PI/3, xtra_xbase + ct, text_ybase);
						g.drawString(o.getName(), xtra_xbase + ct, text_ybase);
						g.rotate(-Math.PI/3, xtra_xbase + ct, text_ybase);
						ct++;
					}
					
					// draw microRNA names
					ct=0;
					for (Gene mir : MirRegList) {
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.PLAIN, 1));
						g.rotate(Math.PI/3, xtra_xbase + this.GOsuper.size() + 2 + ct, text_ybase);
						g.drawString(mir.name, xtra_xbase + this.GOsuper.size() + 2 + ct, text_ybase);
						g.rotate(-Math.PI/3, xtra_xbase + this.GOsuper.size() + 2 + ct, text_ybase);
						ct++;
					}
				}
				
				// draw black frame around module genes expression values
				//g.setStroke(new BasicStroke(0.2F));
				//g.setPaint(Color.black);
				//g.drawRect(x, ybase, node.rightChild.leafDistribution.condSet.size(), this.genes.size()*2);
				//g.drawRect(x, reg_ybase, node.rightChild.leafDistribution.condSet.size(), this.topRegulators.size()*2);
				
				// draw condition labels
				for (int c=0;c<node.rightChild.leafDistribution.condSet.size();c++) {
					int num = node.rightChild.leafDistribution.condSet.get(c);
					String name = this.moduleNetwork.conditionSet.get(num).name;
					Color col = this.moduleNetwork.conditionSet.get(num).col;
					// draw rectangle for tissue type
					g.setPaint(col);
					g.fillRect(x + c, text_ybase - 3, 1, 1);
					g.drawRect(x + c, text_ybase - 3, 1, 1);
					if (this.moduleNetwork.conditionSet.get(num).state.equals("normal")) {
						g.setPaint(Color.lightGray);
						g.fillRect(x + c, text_ybase - 2, 1, 1);
						g.drawRect(x + c, text_ybase - 2, 1, 1);
					}
					else {
						g.setPaint(Color.darkGray);
						g.fillRect(x + c, text_ybase - 2, 1, 1);
						g.drawRect(x + c, text_ybase - 2, 1, 1);
					}
					//g.setPaint(Color.black);
					//g.setStroke(new BasicStroke(0.1F));
					//g.drawRect(x + c, text_ybase - 3, 1, 1);
					//g.drawRect(x + c, text_ybase - 2, 1, 1);
					// rotate canvas and draw string 45/90 deg.
					g.setPaint(Color.black);
					g.setFont(new Font("SansSerif", Font.PLAIN, 1));
					g.rotate(Math.PI/3, x + c, text_ybase);
					g.drawString(name, x + c, text_ybase);
					g.rotate(-Math.PI/3, x + c, text_ybase);
				}
			} 
			// end rightchild *************************************************

			// draw horizontal line between start coordinates for the children
			// = horizontal lines for the tree
			g.setColor(Color.black);
			g.setStroke(extraStr);
			g.drawLine(xl, y + h, xr, y + h);

			// draw the children, right one is drawn lower then left one for clarity
			drawNodeMir(node.leftChild, xl, y + h, h, ybase, g);
			drawNodeMir(node.rightChild, xr, y + h, h, ybase, g);
			// draw line between leaves
			g.setStroke(extraStr);
			g.setPaint(Color.magenta);
			//g.setPaint(new Color(1.0F, 0.0F, 0.5F));
			g.drawLine(x, ybase, x, ybase + this.genes.size()*2);
			g.drawLine(x, reg_ybase, x, reg_ybase + this.topRegulators.size()*2);
			//g.setPaint(Color.black);
			//g.setStroke(new BasicStroke(0.5F));
		}
	}

	/**
	 * Draws regulatory tree leaves with top regulators and GO categories
	 * 
	 * @param node TreeNode to be drawn.
	 * @param x <em>x</em>-coordinate of start position for vertical node line.
	 * @param y <em>y</em>-coordinate of start position for vertical node line.
	 * @param h height of vertical node line
	 * @param ybase <em>y</em>-coordinate of baseline where leaves begin.
	 * @param g graphics in which node is drawn.
	 * 
	 * @author erbon
	 */
	public void drawNodeGO(TreeNode node, int x, int y, int h, int ybase,	Graphics2D g) {
		double[][] data = this.moduleNetwork.data; 
		if (node.nodeStatus.equals("internal")) {
			// where to start drawing regulator expression values
			int x1 = x - node.leftChild.leafDistribution.condSet.size();
			Stroke extraStr = new BasicStroke(1.0F);

//			// draw vertical split line
			g.setPaint(Color.black);
			g.setStroke(new BasicStroke(1));
			g.drawLine(x, y, x, y + h);

			// define y base for top regulators and condition labels
			int reg_ybase = (ybase - this.topRegulators.size() * 2 - 2);
			int text_ybase = ybase + this.genes.size() * 2 + 4;
			
			// leftChild ******************************************************
			int xl;
			if (node.leftChild.nodeStatus.equals("internal")) {
				// if child is also internal, start should be in the middle of its own children
				xl = x - node.leftChild.rightChild.leafDistribution.condSet.size();
			} else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				// = vertical line of the reg tree
				xl = x - node.leftChild.leafDistribution.condSet.size() / 2;
				g.setPaint(Color.black);
				g.setStroke(extraStr);
				g.drawLine(xl, y + h, xl, reg_ybase);
				int xl1 = x - node.leftChild.leafDistribution.condSet.size();
				int gn = 0;
				// draw the data points
				Collections.sort(node.leftChild.leafDistribution.condSet);
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							//Color col = data2Color(gene, data[gene.number][exp]);
							Color col = data2Color(data[gene.number][exp], this.mean, this.sigma);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, ybase + 2*gn, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(xl1 + m, ybase + 2*gn, 1, 2);
					}
					gn++;
				}
				
				// draw top regulators
				int ct=0;
				for (Gene reg : this.topRegulators) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[reg.number][exp])) {
							// convert data to color
							//Color col = data2Color(reg, data[reg.number][exp]);
							Color col = data2Color(data[reg.number][exp], reg.mean, reg.sigma);
							//******************************************************
							//   Option: plot value corresponding to how frequently 
							//   the regulator is assigned to a condition
							//******************************************************
							//Color col = data2Color(this.regConditions.get(reg).get(exp), 0.5, 0.1);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, reg_ybase + 2*ct, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(xl1 + m, reg_ybase + 2*ct, 1, 2);
					}
					ct++;
				}
				
				// draw condition labels
				for (int c=0;c<node.leftChild.leafDistribution.condSet.size();c++) {
					int num = node.leftChild.leafDistribution.condSet.get(c);
					String name = this.moduleNetwork.conditionSet.get(num).name;
//					Color col = this.moduleNetwork.conditionSet.get(num).col;
//					// draw rectangle for tissue type
//					g.setPaint(col);
//					g.fillRect(xl1 + c, text_ybase - 3, 1, 1);
//					g.drawRect(xl1 + c, text_ybase - 3, 1, 1);
//					if (this.moduleNetwork.conditionSet.get(num).state.equals("normal")) {
//						g.setPaint(Color.lightGray);
//						g.fillRect(xl1 + c, text_ybase - 2, 1, 1);
//						g.drawRect(xl1 + c, text_ybase - 2, 1, 1);
//					}
//					else {
//						g.setPaint(Color.darkGray);
//						g.fillRect(xl1 + c, text_ybase - 2, 1, 1);
//						g.drawRect(xl1 + c, text_ybase - 2, 1, 1);
//					}
					// rotate canvas and draw string 45 deg.
					g.setPaint(Color.black);
					g.setFont(new Font("SansSerif", Font.PLAIN, 1));
					g.rotate(Math.PI/3, xl1 + c, text_ybase);
					g.drawString(name, xl1 + c, text_ybase);
					g.rotate(-Math.PI/3, xl1 + c, text_ybase);
				}
			}
			// end leftchild **************************************************

			// rightChild *****************************************************
			int xr;
			float MaxGeneNameLen = 0;
			float GeneNameLen=0;
			g.setFont(new Font("SansSerif", Font.BOLD, 2));
			FontRenderContext frc = g.getFontRenderContext();
			
			if (node.rightChild.nodeStatus.equals("internal"))
				// if child is also internal, start should be in the middle of its own children
				xr = x + node.rightChild.leftChild.leafDistribution.condSet.size();
			else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				// = vertical line of the reg tree
				xr = x + node.rightChild.leafDistribution.condSet.size() / 2;
				g.setPaint(Color.black);
				g.setStroke(extraStr);
				g.drawLine(xr, y + h, xr, reg_ybase);
				int gn = 0;
				// draw the data points
				Collections.sort(node.rightChild.leafDistribution.condSet);
				for (Gene gene : this.genes) {
					GeneNameLen =  (float) g.getFont().getStringBounds(gene.name, frc).getWidth();
					if (GeneNameLen > MaxGeneNameLen)
						MaxGeneNameLen = GeneNameLen;
					for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							//Color col = data2Color(gene, data[gene.number][exp]);
							Color col = data2Color(data[gene.number][exp], this.mean, this.sigma);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, ybase + 2*gn, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(x+m, ybase + 2*gn, 1, 2);
					}
					gn++;
				}
				
				// draw top regulators expression and names
				int ct=0;
				for (Gene reg : this.topRegulators) {
					GeneNameLen =  (float) g.getFont().getStringBounds(reg.name, frc).getWidth();
					if (GeneNameLen > MaxGeneNameLen)
						MaxGeneNameLen = GeneNameLen;
					for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[reg.number][exp])) {
							// convert data to color
							//Color col = data2Color(reg, data[reg.number][exp]);
							Color col = data2Color(data[reg.number][exp], reg.mean, reg.sigma);
							//*****************************************************
							//   Option: plot value corresponding to how frequently 
							//   the regulator is assigned to a condition
							//*****************************************************
							//Color col = data2Color(this.regConditions.get(reg).get(exp), 0.5, 0.1);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, reg_ybase + 2*ct, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(x+m, reg_ybase + 2*ct, 1, 2);
					}
					ct++;
				}
				
				// draw extra information
				if (node == this.hierarchicalTree.rightMostNode()) {
					
					// define extra information x coord. base
					int xtra_xbase = x + node.rightChild.leafDistribution.condSet.size() + (int) MaxGeneNameLen + 5;
					ct=0;
					for (Gene reg : this.topRegulators) {
						
						// draw top reg names
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						g.drawString(reg.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(reg_ybase + 2*ct+2));
						
						ct++;
					}
					ct=0;
					for (Gene gene : this.genes) {
						// draw gene name
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						g.drawString(gene.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(ybase + 2*ct+2));
						
						//draw GO categories
						int ctg = 0;
						for (OntologyTerm o : this.GOsuper) {
							if (gene.GOcat.contains(o)) {
								g.setPaint(Color.orange);
								g.fillRect(xtra_xbase + ctg, ybase + 2*ct,1,2);
								g.setPaint(Color.black);
								g.setStroke(new BasicStroke(0.1F));
								g.drawRect(xtra_xbase + ctg , ybase + 2*ct,1,2);
							}
							else {
								g.setPaint(Color.white);
								g.fillRect(xtra_xbase + ctg, ybase + 2*ct,1,2);
								g.setPaint(Color.black);
								g.setStroke(new BasicStroke(0.1F));
								g.drawRect(xtra_xbase + ctg, ybase + 2*ct,1,2);
							}
							ctg++;
						}
						ct++;
					}
					
					// draw GO labels
					ct=0;
					for (OntologyTerm o : this.GOsuper) {
						// rotate canvas and draw string 45/90 deg.
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.PLAIN, 1));
						g.rotate(Math.PI/3, xtra_xbase + ct, text_ybase);
						g.drawString(o.getName(), xtra_xbase + ct, text_ybase);
						g.rotate(-Math.PI/3, xtra_xbase + ct, text_ybase);
						ct++;
					}
				}					
				
				// draw condition labels
				for (int c=0;c<node.rightChild.leafDistribution.condSet.size();c++) {
					int num = node.rightChild.leafDistribution.condSet.get(c);
					String name = this.moduleNetwork.conditionSet.get(num).name;
//					Color col = this.moduleNetwork.conditionSet.get(num).col;
//					// draw rectangle for tissue type
//					g.setPaint(col);
//					g.fillRect(x + c, text_ybase - 3, 1, 1);
//					g.drawRect(x + c, text_ybase - 3, 1, 1);
//					if (this.moduleNetwork.conditionSet.get(num).state.equals("normal")) {
//						g.setPaint(Color.lightGray);
//						g.fillRect(x + c, text_ybase - 2, 1, 1);
//						g.drawRect(x + c, text_ybase - 2, 1, 1);
//					}
//					else {
//						g.setPaint(Color.darkGray);
//						g.fillRect(x + c, text_ybase - 2, 1, 1);
//						g.drawRect(x + c, text_ybase - 2, 1, 1);
//					}
					// rotate canvas and draw string 45/90 deg.
					g.setPaint(Color.black);
					g.setFont(new Font("SansSerif", Font.PLAIN, 1));
					g.rotate(Math.PI/3, x + c, text_ybase);
					g.drawString(name, x + c, text_ybase);
					g.rotate(-Math.PI/3, x + c, text_ybase);
				}
			} 
			// end rightchild *************************************************

			// draw horizontal line between start coordinates for the children
			// = horizontal lines for the tree
			g.setColor(Color.black);
			g.setStroke(extraStr);
			g.drawLine(xl, y + h, xr, y + h);

			// draw the children, right one is drawn lower then left one for clarity
			drawNodeGO(node.leftChild, xl, y + h, h, ybase, g);
			drawNodeGO(node.rightChild, xr, y + h, h, ybase, g);
			// draw line between leaves
			g.setStroke(extraStr);
			g.setPaint(Color.magenta);
			g.drawLine(x, ybase, x, ybase + this.genes.size()*2);
			g.drawLine(x, reg_ybase, x, reg_ybase + this.topRegulators.size()*2);
		}
	}



	/**
	 * Draws regulatory tree leaves with top regulators and tree, Eps format
	 * 
	 * @param node TreeNode to be drawn.
	 * @param x <em>x</em>-coordinate of start position for vertical node line.
	 * @param y <em>y</em>-coordinate of start position for vertical node line.
	 * @param h height of vertical node line
	 * @param ybase <em>y</em>-coordinate of baseline where leaves begin.
	 * @param g graphics in which node is drawn.
	 * 
	 * @author erbon
	 */
	public void drawNodeEps(TreeNode node, int x, int y, int h, int ybase,	Graphics2D g) {
		double[][] data = this.moduleNetwork.data; 
		if (node.nodeStatus.equals("internal")) {
			// where to start drawing regulator expression values
			//int x1 = x - node.leftChild.leafDistribution.condSet.size();
			Stroke extraStr = new BasicStroke(1.0F);

			// draw vertical split line
			g.setPaint(Color.black);
			g.setStroke(new BasicStroke(1));
			g.drawLine(x, y, x, y + h);

			// define y base for top regulators and condition labels
			int reg_ybase = (ybase - this.topRegulators.size() * 2 - 2);
			int text_ybase = ybase + this.genes.size() * 2 + 4;
			
			// leftChild ******************************************************
			int xl;
			if (node.leftChild.nodeStatus.equals("internal")) {
				// if child is also internal, start should be in the middle of its own children
				xl = x - node.leftChild.rightChild.leafDistribution.condSet.size();
			} else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				// = vertical line of the reg tree
				xl = x - node.leftChild.leafDistribution.condSet.size() / 2;
				g.setPaint(Color.black);
				g.setStroke(extraStr);
				g.drawLine(xl, y + h, xl, reg_ybase);
				int xl1 = x - node.leftChild.leafDistribution.condSet.size();
				int gn = 0;
				// draw the data points
				Collections.sort(node.leftChild.leafDistribution.condSet);
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							if (this.moduleNetwork.getGlobalMeanForFigures() == false) {
								Color col = data2Color(data[gene.number][exp], this.mean, this.sigma);
								g.setPaint(col);
							}
							else {
								Color col = data2Color(data[gene.number][exp], this.moduleNetwork.dataMean, this.moduleNetwork.dataSigma);
								g.setPaint(col);
							}
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, ybase + 2*gn, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(xl1 + m, ybase + 2*gn, 1, 2);
					}
					gn++;
				}
				
				// draw top regulators
				int ct=0;
				for (Gene reg : this.topRegulators) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[reg.number][exp])) {
							// convert data to color
							if (this.moduleNetwork.getRegulatorMeanForFigures() == true) {
								Color col = data2Color(data[reg.number][exp], reg.mean, reg.sigma);
								g.setPaint(col);
							}
							else if (this.moduleNetwork.getGlobalMeanForFigures() == false) {
								Color col = data2Color(data[reg.number][exp], this.mean, this.sigma);
								g.setPaint(col);
							}
							else {
								Color col = data2Color(data[reg.number][exp], this.moduleNetwork.dataMean, this.moduleNetwork.dataSigma);
								g.setPaint(col);
							}
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, reg_ybase + 2*ct, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(xl1 + m, reg_ybase + 2*ct, 1, 2);
					}
					ct++;
				}
				
				// draw condition labels
				for (int c=0;c<node.leftChild.leafDistribution.condSet.size();c++) {
					int num = node.leftChild.leafDistribution.condSet.get(c);
					String name = this.moduleNetwork.conditionSet.get(num).name;
					// rotate canvas and draw string 45 deg.
					g.setPaint(Color.black);
					g.setFont(new Font("SansSerif", Font.PLAIN, 1));
					g.rotate(Math.PI/2, xl1 + c, text_ybase);
					g.drawString(name, xl1 + c, text_ybase);
					g.rotate(-Math.PI/2, xl1 + c, text_ybase);
				}
			}
			// end leftchild **************************************************

			// rightChild *****************************************************
			int xr;
			float MaxGeneNameLen = 0;
			float GeneNameLen=0;
			g.setFont(new Font("SansSerif", Font.BOLD, 2));
			FontRenderContext frc = g.getFontRenderContext();
			
			if (node.rightChild.nodeStatus.equals("internal"))
				// if child is also internal, start should be in the middle of its own children
				xr = x + node.rightChild.leftChild.leafDistribution.condSet.size();
			else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				// = vertical line of the reg tree
				xr = x + node.rightChild.leafDistribution.condSet.size() / 2;
				g.setPaint(Color.black);
				g.setStroke(extraStr);
				g.drawLine(xr, y + h, xr, reg_ybase);
				int gn = 0;
				// draw the data points
				Collections.sort(node.rightChild.leafDistribution.condSet);
				for (Gene gene : this.genes) {
					GeneNameLen =  (float) g.getFont().getStringBounds(gene.name, frc).getWidth();
					if (GeneNameLen > MaxGeneNameLen)
						MaxGeneNameLen = GeneNameLen;
					for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							if (this.moduleNetwork.getGlobalMeanForFigures() == false) {
								Color col = data2Color(data[gene.number][exp], this.mean, this.sigma);
								g.setPaint(col);
							}
							else {
								Color col = data2Color(data[gene.number][exp], this.moduleNetwork.dataMean, this.moduleNetwork.dataSigma);
								g.setPaint(col);
							}
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, ybase + 2*gn, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(x+m, ybase + 2*gn, 1, 2);
					}
					gn++;
				}
				
				// draw top regulators expression and names
				int ct=0;
				for (Gene reg : this.topRegulators) {
					GeneNameLen =  (float) g.getFont().getStringBounds(reg.name, frc).getWidth();
					if (GeneNameLen > MaxGeneNameLen)
						MaxGeneNameLen = GeneNameLen;
					for (int m = 0; m < node.rightChild.leafDistribution.condSet.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[reg.number][exp])) {
							// convert data to color
							if (this.moduleNetwork.getRegulatorMeanForFigures() == true) {
								Color col = data2Color(data[reg.number][exp], reg.mean, reg.sigma);
								g.setPaint(col);
							}
							else if (this.moduleNetwork.getGlobalMeanForFigures() == false) {
								Color col = data2Color(data[reg.number][exp], this.mean, this.sigma);
								g.setPaint(col);
							}
							else {
								Color col = data2Color(data[reg.number][exp], this.moduleNetwork.dataMean, this.moduleNetwork.dataSigma);
								g.setPaint(col);
							}
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, reg_ybase + 2*ct, 1, 2);
						g.setPaint(Color.black);
						g.setStroke(new BasicStroke(0.1F));
						g.drawRect(x+m, reg_ybase + 2*ct, 1, 2);
					}
					ct++;
				}
				
				// draw extra information
				if (node == this.hierarchicalTree.rightMostNode()) {
					
					// define extra information x coord. base
					ct=0;
					for (Gene reg : this.topRegulators) {
						
						// draw top reg names
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						g.drawString(reg.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(reg_ybase + 2*ct+2));
						
						ct++;
					}
					ct=0;
					for (Gene gene : this.genes) {
						// draw gene name
						g.setPaint(Color.black);
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						g.drawString(gene.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(ybase + 2*ct+2));
						
						ct++;
					}
					
				}					
				
				// draw condition labels
				for (int c=0;c<node.rightChild.leafDistribution.condSet.size();c++) {
					int num = node.rightChild.leafDistribution.condSet.get(c);
					String name = this.moduleNetwork.conditionSet.get(num).name;
					// rotate canvas and draw string 45/90 deg.
					g.setPaint(Color.black);
					g.setFont(new Font("SansSerif", Font.PLAIN, 1));
					g.rotate(Math.PI/2, x + c, text_ybase);
					g.drawString(name, x + c, text_ybase);
					g.rotate(-Math.PI/2, x + c, text_ybase);
				}
			} 
			// end rightchild *************************************************

			// draw horizontal line between start coordinates for the children
			// = horizontal lines for the tree
			g.setColor(Color.black);
			g.setStroke(extraStr);
			g.drawLine(xl, y + h, xr, y + h);

			// draw the children, right one is drawn lower then left one for clarity
			drawNodeEps(node.leftChild, xl, y + h, h, ybase, g);
			drawNodeEps(node.rightChild, xr, y + h, h, ybase, g);
			// draw line between leaves
			g.setStroke(extraStr);
			g.setPaint(Color.magenta);
			g.drawLine(x, ybase, x, ybase + this.genes.size()*2);
			g.drawLine(x, reg_ybase, x, reg_ybase + this.topRegulators.size()*2);
		}
	}

	
	/**
	 * Draws the expression values of this module as one big leaf. Used 
	 * for modules without parents.
	 * 
	 * @param x <em>x</em>-coordinate of start position for drawing.
	 * @param y <em>y</em>-coordinate of start position for drawing.
	 * @param g graphics in which node is drawn.
	 * 
	 * @author tomic
	 */
	public void draw1leaf(int x, int y, Graphics2D g) {
		double[][] data = this.moduleNetwork.data;
		int numCond = this.hierarchicalTree.leafDistribution.condSet.size();
		int gn = 0;
		for (Gene gene : this.genes) {
			for (int m : this.hierarchicalTree.leafDistribution.condSet) {
				if (!Double.isNaN(data[gene.number][m])) {
					// convert data to color
					Color col = data2Color(gene, data[gene.number][m]);
					g.setPaint(col);
				} else
					// missing value
					g.setPaint(Color.white);
				g.fillRect(x + m, y + 2*gn, 1, 2);
			}
			//add name of gene
			g.setFont(new Font("SansSerif", Font.BOLD, 2));
			g.setPaint(Color.black);
			g.drawString(gene.name, new Float(x + numCond+1), new Float(y + 2*gn +2));
			gn++;
		}
		g.setPaint(Color.black);
		// draw frame around expression values
		g.drawRect(x, y, numCond, this.genes.size()*2);
	}
	

	/**
	 * Draws the expression values of this module as one big leaf. Used 
	 * for modules without parents, Eps format.
	 * 
	 * @param x <em>x</em>-coordinate of start position for drawing.
	 * @param y <em>y</em>-coordinate of start position for drawing.
	 * @param g graphics in which node is drawn.
	 * 
	 * @author erbon
	 */
	public void drawOneLeafEps(int x, int y, Graphics2D g) {
		double[][] data = this.moduleNetwork.data;
		int numCond = this.hierarchicalTree.leafDistribution.condSet.size();
		int gn = 0;
		
		// Define coordinate to draw conditions' labels
		int text_ybase = y + this.genes.size() * 2 + 4;
		
		for (Gene gene : this.genes) {
			for (int m : this.hierarchicalTree.leafDistribution.condSet) {
				if (!Double.isNaN(data[gene.number][m])) {
					// convert data to color
					//Color col = data2Color(gene, data[gene.number][m]);
					if (this.moduleNetwork.getGlobalMeanForFigures() == false) {
						Color col = data2Color(data[gene.number][m], this.mean, this.sigma);
						g.setPaint(col);
					}
					else {
						Color col = data2Color(data[gene.number][m], this.moduleNetwork.dataMean, this.moduleNetwork.dataSigma);
						g.setPaint(col);
					}
				} else
					// missing value
					g.setPaint(Color.white);
				g.fillRect(x + m, y + 2*gn, 1, 2);
				g.setPaint(Color.black);
				g.setStroke(new BasicStroke(0.1F));
				g.drawRect(x + m, y + 2*gn, 1, 2);
			}
			//add name of gene
			g.setFont(new Font("SansSerif", Font.BOLD, 2));
			g.setPaint(Color.black);
			g.drawString(gene.name, new Float(x + numCond+1), new Float(y + 2*gn + 2));
			gn++;
		}
		
		// draw condition labels
		for (int c=0; c < this.hierarchicalTree.leafDistribution.condSet.size();c++) {
			int num = this.hierarchicalTree.leafDistribution.condSet.get(c);
			String name = this.moduleNetwork.conditionSet.get(num).name;
			// rotate canvas and draw string 90 deg.
			g.setPaint(Color.black);
			g.setFont(new Font("SansSerif", Font.PLAIN, 1));
			g.rotate(Math.PI/2, x + c, text_ybase);
			g.drawString(name, x + c, text_ybase);
			g.rotate(-Math.PI/2, x + c, text_ybase);
		}
		
		g.setPaint(Color.black);
		// draw frame around expression values
		g.drawRect(x, y, numCond, this.genes.size()*2);
	}
	
	/**
	 * Draws the expression values of this module as one big leaf. Used 
	 * for modules without parents.
	 * 
	 * @param x <em>x</em>-coordinate of start position for drawing.
	 * @param y <em>y</em>-coordinate of start position for drawing.
	 * @param g graphics in which node is drawn.
	 * 
	 * @author tomic
	 */
	public void draw1leafExtra(ArrayList<Gene> geneSet, int x, int y, Graphics2D g) {
		double[][] data = this.moduleNetwork.data;
		int numCond = this.hierarchicalTree.leafDistribution.condSet.size();
		int gn = 0;
		for (Gene gene : this.genes) {
			for (int m : this.hierarchicalTree.leafDistribution.condSet) {
				if (!Double.isNaN(data[gene.number][m])) {
					// convert data to color
					Color col = data2Color(gene, data[gene.number][m]);
					g.setPaint(col);
				} else
					// missing value
					g.setPaint(Color.white);
				g.fillRect(x + m, y + 2*gn, 1, 2);
			}
			//add name of gene
			g.setFont(new Font("SansSerif", Font.BOLD, 2));
			if (geneSet.contains(gene))
				g.setPaint(Color.red);
			else
				g.setPaint(Color.black);
			g.drawString(gene.name, new Float(x + numCond+1), new Float(y + 2*gn +2));
			gn++;
		}
		g.setPaint(Color.black);
		// draw frame around expression values
		g.drawRect(x, y, numCond, this.genes.size()*2);
	}

	/**
	 * Draws the expression values below a specific tree node as one leaf. 
	 * @param node
	 * @param x
	 * @param y
	 * @param g
	 * 
	 * @author tomic
	 */
	public void drawLeaf(TreeNode node, int x, int y, Graphics2D g){
		double[][] data = this.moduleNetwork.data;
		int numCond = node.leafDistribution.condSet.size();
		int gn = 0;
		for (Gene gene : this.genes) {
			for (int m=0;m<node.leafDistribution.condSet.size(); m++) {
				int exp = node.leafDistribution.condSet.get(m);
				if (!Double.isNaN(data[gene.number][exp])) {
					// convert data to color
					Color col = data2Color(gene, data[gene.number][exp]);
					g.setPaint(col);
				} else
					// missing value
					g.setPaint(Color.white);
				g.fillRect(x + m, y + 2*gn, 1, 2);
			}
			gn++;
		}
		g.setPaint(Color.black);
		// draw frame around expression values
		g.drawRect(x, y, numCond, this.genes.size()*2);
	}

	/**
	 * Same as {@link Module#draw(String)}, except no regulators are drawn at the tree nodes.
	 * 
	 * @see Module#draw(String)
	 * @param filename
	 * 
	 * @author tomic
	 */
	public void drawNoReg(String filename) {
		// scale factor
		int sc = 4;
		// coordinates of origin
		int x0 = 10;
		int y0 = 10;
		// tree step height
		int h = 18;
		// figure width
		int width = 3 * x0 + this.hierarchicalTree.leafDistribution.condSet.size();
		int height;
		int x = x0;
		int ybase = y0;
		
		if (this.hierarchicalTree.nodeStatus.equals("leaf")){ // no tree
			height = 2 * y0 + this.genes.size()*2;
		}	
		else {	
			// coordinates of root
			x = x0 + this.hierarchicalTree.leftChild.leafDistribution.condSet.size();
			ybase = y0 + (h + 4) * (this.hierarchicalTree.treeDepth() + 1);
			height = ybase + this.genes.size()*2 + y0;
		}
		BufferedImage image = new BufferedImage(sc * width, sc * height,
				BufferedImage.TYPE_INT_BGR);
		Graphics2D g2;
		g2 = image.createGraphics();
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		g2.scale(sc, sc);
		g2.setStroke(new BasicStroke(0.5F));
		if (this.hierarchicalTree.nodeStatus.equals("leaf")) { // no tree
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			// draw data
			draw1leaf(x0, y0, g2);
		} else {
			g2.setPaint(Color.white);
			g2.fillRect(0, 0, width, height);
			// draw regression tree
			drawnodeNoReg(this.hierarchicalTree, x, y0, h, ybase, g2);
		}
		File outFile = new File(filename);
		try {
			ImageIO.write(image, "png", outFile);
		} catch (IOException e) {
			return;
		}
	}

	/**
	 * Same as {@link Module#drawnode(TreeNode, int, int, int, int, Graphics2D)} except no 
	 * regulator is drawn at the node
	 * 
	 * @see Module#drawnode(TreeNode, int, int, int, int, Graphics2D)
	 * @author tomic
	 */
	public void drawnodeNoReg(TreeNode node, int x, int y, int h, int ybase, Graphics2D g) {
		// input:
		// (x,y): start coordinates, h: height of vertical node line,
		// ybase: baseline where leaves begin
		// g: graphics in which node is drawn
		double[][] data = this.moduleNetwork.data;
		if (node.nodeStatus.equals("internal")) {
			// where to start drawing regulator expression values
			int x1 = x - node.leftChild.leafDistribution.condSet.size();
			g.setPaint(Color.black);
			// draw vertical split line
			g.drawLine(x, y, x, y + h);
			// annotate split with its score gain
			double diff = (-node.leafDistribution.score
					+ node.leftChild.leafDistribution.score + node.rightChild.leafDistribution.score)
					/ this.moduleNetwork.numGenes;
			String quality = String.format("%.2g", diff);
			g.setFont(new Font("SansSerif", Font.BOLD, 3));
			g.drawString(quality, new Float(x), new Float(y + h + 3));
			// x-coordinate for leftChild
			int xl;
			if (node.leftChild.nodeStatus.equals("internal")) {
				// if child is also internal, start should be in the middle of
				// its own children
				xl = x - node.leftChild.rightChild.leafDistribution.condSet.size();
			} else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				xl = x - node.leftChild.leafDistribution.condSet.size() / 2;
				g.drawLine(xl, y + h, xl, ybase);
				int xl1 = x - node.leftChild.leafDistribution.condSet.size();
				int gn = 0;
				// draw the data points
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.leftChild.leafDistribution.condSet.size(); m++) {
						int exp = node.leftChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							Color col = data2Color(gene, data[gene.number][exp]);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(xl1 + m, ybase + 2*gn, 1, 2);
					}
					gn++;
				}
				g.setPaint(Color.black);
				// draw frame around expression values
				g.drawRect(xl1, ybase, node.leftChild.leafDistribution.condSet.size(), this.genes.size()*2);
				// print average posterior probability of leaf
				g.setFont(new Font("SansSerif", Font.BOLD, 4));
				double p = node.leftChild.leafDistribution.gaussianParam[1];
				String scr =  String.format("%.2f", p);
				FontRenderContext frc = g.getFontRenderContext();
				LineMetrics metrics = g.getFont().getLineMetrics(scr, frc);
				float scrw = (float) g.getFont().getStringBounds(scr, frc).getWidth();
				g.drawString(scr, new Float(xl-0.5*scrw), new Float(ybase+2*this.genes.size()+5));
				scr =  String.format("%.2f", node.leftChild.leafDistribution.gaussianParam[0]);
				metrics = g.getFont().getLineMetrics(scr, frc);
				scrw = (float) g.getFont().getStringBounds(scr, frc).getWidth();
				g.drawString(scr, new Float(xl-0.5*scrw), new Float(ybase+2*this.genes.size()+9.5));
			}
			// x-coordinate for rightChild
			int xr;
			if (node.rightChild.nodeStatus.equals("internal"))
				// if child is also internal, start should be in the middle of
				// its own children
				xr = x + node.rightChild.leftChild.leafDistribution.condSet.size();
			else {
				// if child is leaf, we draw a vertical to the middle of the
				// leaf as well as the leaf data point
				xr = x + node.rightChild.leafDistribution.condSet.size() / 2;
				g.drawLine(xr, y + h, xr, ybase);
				int gn = 0;
				// draw the data points
				for (Gene gene : this.genes) {
					for (int m = 0; m < node.rightChild.leafDistribution.condSet
					.size(); m++) {
						int exp = node.rightChild.leafDistribution.condSet.get(m);
						if (!Double.isNaN(data[gene.number][exp])) {
							// convert data to color
							Color col = data2Color(gene, data[gene.number][exp]);
							g.setPaint(col);
						} else
							// missing value
							g.setPaint(Color.white);
						g.fillRect(x + m, ybase + 2*gn, 1, 2);
					}
					if(node == this.hierarchicalTree.rightMostNode()){
						//add names of genes to plot
						g.setFont(new Font("SansSerif", Font.BOLD, 2));
						g.setPaint(Color.black);
						g.drawString(gene.name, new Float(x + node.rightChild.leafDistribution.condSet.size() +1), new Float(ybase + 2*gn+2));
					}
					gn++;
				}
				g.setPaint(Color.black);
				// draw frame around expression values
				g.drawRect(x, ybase, node.rightChild.leafDistribution.condSet.size(), this.genes.size()*2);
				// print average posterior probability of leaf
				g.setFont(new Font("SansSerif", Font.BOLD, 4));
				double p = node.leftChild.leafDistribution.gaussianParam[1];
				String scr =  String.format("%.2f", p);
				FontRenderContext frc = g.getFontRenderContext();
				LineMetrics	metrics = g.getFont().getLineMetrics(scr, frc);
				float scrw = (float) g.getFont().getStringBounds(scr, frc).getWidth();
				g.drawString(scr, new Float(xr-0.5*scrw), new Float(ybase+2*this.genes.size()+5));
				
				scr =  String.format("%.2f", node.rightChild.leafDistribution.gaussianParam[0]);
				metrics = g.getFont().getLineMetrics(scr, frc);
				scrw = (float) g.getFont().getStringBounds(scr, frc).getWidth();
				g.drawString(scr, new Float(xr-0.5*scrw), new Float(ybase+2*this.genes.size()+9.5));
			}
			// draw horizontal line between start coordinates for the children
			g.drawLine(xl, y + h, xr, y + h);
			// draw the children, right one is drawn lower then left one for clarity
			drawnodeNoReg(node.leftChild, xl, y + h, h, ybase, g);
			drawnodeNoReg(node.rightChild, xr, y + h, h + 4, ybase, g);
			// draw line between leaves
			g.setPaint(Color.red);
			g.drawLine(x, ybase, x, ybase + this.genes.size()*2);
			g.setPaint(Color.black);
		}
	}

	/**
	 * Learn a regulation tree top-down using exact splits on a given set of
	 * parent for this module.
	 * 
	 * @author tomic
	 *
	 */
	public void learnTreeTopDown(ArrayList<Gene> regs, int maxParents){
		System.out.println("Learning tree for module " + this.number);
		// reset tree and recompute statistics
		this.parents = new ArrayList<Gene>();
		this.hierarchicalTrees = new ArrayList<TreeNode>();
		this.hierarchicalTree = new TreeNode(this, this.moduleNetwork.normalGammaPrior);
		for (int m = 0; m < this.moduleNetwork.data[0].length; m++)
			this.hierarchicalTree.leafDistribution.condSet.add(m);
		this.hierarchicalTree.statistics();
		// update score value
		this.moduleScore = this.hierarchicalTree.bayesianScore();
		// keep splitting as long as we have parents and can improve the score
		//ArrayList<Gene> regs = this.parents;
		boolean foundSplit = true;
		while (foundSplit && this.parents.size() < maxParents){
			// find current best split
			foundSplit = this.bestSplitTopDown(regs);
			// apply best split
			if (foundSplit){
				System.out.println("\t found split: " + this.bestSplit.regulator.name);
				this.bestSplit.node.split(this.bestSplit.regulator, this.bestSplit.splitValue);
				this.parents.add(this.bestSplit.regulator);
			}
		}
	}
	
	/**
	 * Finds the best split in a module using exact splits on a subset of regulators.
	 * If one is found it is stored in {@link Module#bestSplit}.
	 * 
	 * @return <code>true</code> if a split is found.
	 * 
	 * @author tomic
	 */
	public boolean bestSplitTopDown(ArrayList<Gene> regulators) {
		HashSet<TreeNode> leaves = new HashSet<TreeNode>();
		TreeNode optLeaf = new TreeNode();
		double optSplit = 0.0;
		Gene optReg = new Gene();
		double optModuleScore;
		double maxModuleScore;
		double scoreGain;

		maxModuleScore = this.moduleScore;

		this.hierarchicalTree.gatherLeaves(leaves);
		for (TreeNode leaf : leaves)
			for (Gene reg : regulators) {
				for (Double splt : reg.possibleSplitValues) {
					// try to split
					optModuleScore = leaf.trySplit(reg, splt);
					if (optModuleScore - maxModuleScore > 0.0) {
						optLeaf = leaf;
						optReg = reg;
						optSplit = splt;
						maxModuleScore = optModuleScore;
					}
				}
			}

		// we have found a best split if max has changed its value in the
		// previous loops
		if (maxModuleScore - this.moduleScore > 0.0) {
			scoreGain = maxModuleScore - this.moduleScore;
			this.bestSplit = new Split(optLeaf, optReg, optSplit, 0.0, scoreGain);
			return true;
		} else {
			this.bestSplit = new Split();
			return false;
		}
	}
	
	/**
	 * Finds the best split in a module which preseves acyclicity. 
	 * If one is found it is stored in {@link Module#bestSplit}.
	 * 
	 * @param maxParents maximum number of parents for a module.
	 * @return <code>true</code> if a split is found.
	 * 
	 * @author tomic
	 */
	public boolean bestSplitTopDown(int maxParents) {
		HashSet<TreeNode> leaves = new HashSet<TreeNode>();
		TreeNode optLeaf = new TreeNode();
		double optSplit = 0.0;
		Gene optReg = new Gene();
		double optModuleScore;
		double maxModuleScore;
		double scoreGain;

		System.out.println("Looking for best split in module " + this.number);
		maxModuleScore = this.moduleScore;

		// only proceed if module has less than maxParents parents
		if (this.parents.size() < maxParents) {
			for (Gene reg : this.moduleNetwork.regulatorSet) {
				// only proceed if adding reg to parents preserves acyclicity
				if (!this.moduleNetwork.moduleGraph.testCycle(reg, this)) {
					this.hierarchicalTree.gatherLeaves(leaves);
					for (TreeNode leaf : leaves)
						for (Double splt : reg.possibleSplitValues) {
							// try to split
							optModuleScore = leaf.trySplit(reg, splt);
							// System.out.println(test);
							if (optModuleScore - maxModuleScore > 0.0) {
								optLeaf = leaf;
								optReg = reg;
								optSplit = splt;
								maxModuleScore = optModuleScore;
							}
						}
				}
			}
		}

		// we have found a best split if max has changed its value in the
		// previous loops
		if (maxModuleScore - this.moduleScore > 0.0) {
			scoreGain = maxModuleScore - this.moduleScore;
			this.bestSplit = new Split(optLeaf, optReg, optSplit, 0.0, scoreGain);
			System.out.println("... best split found with score gain "
					+ scoreGain / this.moduleNetwork.numGenes);
			return true;
		} else {
			this.bestSplit = new Split();
			System.out.println("... no best split found");
			return false;
		}
	}
	

	

	/**
	 * Computes the best split of the merged leaves below a node after a
	 * reassignment.
	 * 
	 * @param node node for which split is searched.
	 * 
	 * @author tomic
	 * 
	 */
	public Split bestMergeSplit(TreeNode node) {
		if (node.nodeStatus == "leaf")
			return new Split();
		// get list of leaves below node
		ArrayList<TreeNode> leafList = new ArrayList<TreeNode>();
		node.gatherLeafList(leafList);
		// create fictitious node to hold all leaves, note that
		// statistics are additive
		TreeNode fictnode = new TreeNode(this, node.leafDistribution.normalGammaPrior);
		for (TreeNode leaf : leafList) {
			fictnode.leafDistribution.statistics[0] += leaf.leafDistribution.statistics[0];
			fictnode.leafDistribution.statistics[1] += leaf.leafDistribution.statistics[1];
			fictnode.leafDistribution.statistics[2] += leaf.leafDistribution.statistics[2];
			for (int m : leaf.leafDistribution.condSet)
				fictnode.leafDistribution.condSet.add(m);
		}
		// compute fictitious node score: this is the unsplitted score
		fictnode.leafDistribution.bayesianScore();
		// find the best split of fictnode, set initially to current
		double optSplit = node.regulationSplit.splitValue;
		Gene optReg = node.regulationSplit.regulator;
		// score of current split
		double scoreGain = fictnode.trySplit(optReg, optSplit);
		double optScoreGain;
		// first try other split values with optReg
		for (Double splt : optReg.possibleSplitValues) {
			if (splt != node.regulationSplit.splitValue) {
				optScoreGain = fictnode.trySplit(optReg, splt);
				if (optScoreGain - scoreGain > 0.0) {
					optSplit = splt;
					scoreGain = optScoreGain;
				}
			}
		}
		// try other regulators that preserve acyclicity
		for (Gene reg : this.moduleNetwork.regulatorSet) {
			if (reg.number != optReg.number
					&& !this.moduleNetwork.moduleGraph.testCycle(reg, this)) {
				for (Double splt : reg.possibleSplitValues) {
					optScoreGain = fictnode.trySplit(reg, splt);
					if (optScoreGain - scoreGain > 0.0) {
						optReg = reg;
						optSplit = splt;
						scoreGain = optScoreGain;
					}
				}
			}
		}

		// there must be a best split, at the very least the current
		// one; note this is a split for fictnode
		return new Split(fictnode, optReg, optSplit, 0.0, scoreGain);
	}
	
	/**
	 * Trims a regulation tree by removing nodes that are no longer best splits
	 * after a reassignment. Resets mod.bestSplit if applicable
	 * 
	 * @param mod Module where trimming is performed.
	 * @param node node below which trimming is performed.
	 * 
	 * @author tomic
	 * 
	 */
	public void treeTrim(TreeNode node) {
		Split splt;
		if (node.nodeStatus == "internal") {
			System.out.println("Testing node " + node.regulationSplit.regulator.name);
			// find the best split
			splt = this.bestMergeSplit(node);
			// check if it is still the same, if yes go down in tree,
			// if not prune the tree, but don't apply the new split
			if (splt.regulator.number == node.regulationSplit.regulator.number
					&& splt.splitValue == node.regulationSplit.splitValue) {
				System.out.println("... keep");
				this.treeTrim(node.leftChild);
				this.treeTrim(node.rightChild);
			} else {
				System.out.println("... trim");
				// remove regulators below and including node from parent set
				// and modulegraph
				removeRegulators(node);
				// make it a leaf, note that all the information is allready in
				// the bestMergeSplit
				node.nodeStatus = "leaf";
				node.leftChild = null;
				node.rightChild = null;
				node.leafDistribution = splt.node.leafDistribution;
				// set splt.leaf to node so later split is applied in the right place
				splt.node= node;
				// check if the split would be a best split for the current module
				if (splt.scoreGain >= this.bestSplit.scoreGain)
					this.bestSplit = splt;
			}
		}
	}
	

	/**
	 * Removes all regulators in a subtree starting at a node.
	 * 
	 * @param node start node.
	 * 
	 * @author tomic
	 */
	public void removeRegulators(TreeNode node) {
		if (node.nodeStatus == "internal") {
			this.moduleNetwork.moduleGraph.removeRegulator(node.regulationSplit.regulator, node.module);
			node.module.parents.remove(node.regulationSplit.regulator);
			removeRegulators(node.leftChild);
			removeRegulators(node.rightChild);
		}
	}

}
