/* LeMoNe - Learning Module Network
 * 
 * Copyright (C) 2005-2008 VIB 
 * 
 * Written by Tom Michoel, Eric Bonnet, Anagha Joshi, Steven Maere
 *  
*/


package be.ugent.psb.ModuleNetwork;

import java.util.*;
import java.io.*;
import java.text.*;
//import be.ugent.psb.startupprops.*;

/**
 * Run one of LeMoNe's network inference algorithms. 
 * @author erbon, tomic
 * 
 */
public class Run {
	/**
	 * Run one of LeMoNe's network inference algorithms. Parameter settings
	 * are read from a properties file, see <code>LeMoNe.properties.new</code>
	 * for an example, and the inferred network (or network ensemble) is stored 
	 * in an XML file. The different algorithms are:
	 * <ol>
	 * <li>No initial clustering, no Gibbs sampling for genes and experiments,
	 * assigning the best regulator to each tree node</li>
	 * <li>No initial clustering, Gibbs sampling for experiments but not for genes,
	 * assigning regulators stochastically or by minimum. First finds optimal gene 
	 * modules by iteratively Gibbs sampling one hierarchical tree per module, and 
	 * reassigning genes. After convergence, generate different experiment partitions 
	 * each with their own regulation tree.</li>
	 * <li>Gibbs sampling for genes and experiments, assigning regulators stochastically
	 * or by minimum. Finds different sets of modules by iteratively Gibbs sampling genes and 
     * experiments. For each set of modules sampled, multiple hierarchical trees
     * can be sampled per module by separately Gibbs sampling experiments.</li>
	 * </ol>
	 *  
	 * @param args name of properties file, uses default file <code>LeMoNe.properties</code> 
	 * if absent
	 * 
	 * @author tomic
	 */
    public static void main(String[] args) {

    	ModuleNetwork M = new ModuleNetwork();
    	
        // get properties in a hashmap
        HashMap<String,String> hp = new HashMap<String,String>();
        
        // ModuleNetwork parameter file is found in the current directory
        // or passed as first argument
        
        //parse commandline
        ArrayList<String> flagsArgs = new ArrayList<String>();
        ArrayList<String> filesArgs = new ArrayList<String>();
        for (String arg : args){
        	if (arg.startsWith("-")){
        		flagsArgs.add(arg);
        	} else {
        		filesArgs.add(arg);
        	}
        }
        boolean useGUI = flagsArgs.contains("--gui");
        // load properties
//		StartUpPropertiesGUI startupGUI = null;
		if (useGUI) {
//			if (filesArgs.size() > 0) {
//				startupGUI = new StartUpPropertiesGUI(
//						new File(filesArgs.get(0)));
//			} else {
//				startupGUI = new StartUpPropertiesGUI(new File(
//						"LeMoNe.properties"));
//			}
//			getProperties(startupGUI.showStartupDialog(100, 150).toString(), hp);
		} else {
			if (filesArgs.size() > 0) {
				getProperties(filesArgs.get(0), hp);
			} else {
				getProperties("LeMoNe.properties", hp);
			}
		}
        
		
        // Set data directory
        String dataDir = setProperty("dataDir",hp);

        // Set expression data file name
        String expressiondata = setProperty("expressiondata",hp);

        // data type
        String dataType = setProperty("dataType",hp);
                
        // Set regulators file name
        String regulatordata = setProperty("regulatordata",hp);

        // Perform initial clustering?
        boolean initClust = Boolean.parseBoolean(setProperty("initClust",hp));

        // Set initial number of clusters and initial cluster file name if necessary
        int numclust = 0;
        String clusterdata = "";
        if (!initClust){
        	numclust = Integer.parseInt(setProperty("numclust",hp));
        	clusterdata = setProperty("clusterdata",hp);
        }

        // Set normal-gamma prior paramaters
        double lambda0 = Double.parseDouble(setProperty("lambda0",hp));
        double mu0 = Double.parseDouble(setProperty("mu0",hp));
        double alpha0 = Double.parseDouble(setProperty("alpha0",hp));
        double beta0 = Double.parseDouble(setProperty("beta0",hp));

        // Use Gibbs sampling or heuristic search to optimize modules?
        boolean gibbsSampleGenes = Boolean.parseBoolean(setProperty("gibbsSampleGenes",hp));

        // If Gibbs sampling genes, set parameters
        int numRunsGenes=0, numStepsGenes=0, burnInGenes=0, sampleStepGenes=0, initNumClust=0;
        double epsConvergence = 0.0;
        if (gibbsSampleGenes) {
        	//initial number of clusters
        	initNumClust = Integer.parseInt(setProperty("initNumClust",hp));
        	// number of runs
        	numRunsGenes = Integer.parseInt(setProperty("numRunsGenes",hp));
        	// number of steps in each run
        	numStepsGenes = Integer.parseInt(setProperty("numStepsGenes",hp));
        	// number of initial steps before sampling starts
        	burnInGenes = Integer.parseInt(setProperty("burnInGenes",hp));
        	// number of steps between samples
        	sampleStepGenes = Integer.parseInt(setProperty("sampleStepGenes",hp));
        } else { // set heuristic search convergence parameter
        	epsConvergence = Double.parseDouble(setProperty("epsConvergence",hp));
        }

        // Use Gibbs sampling or heuristic search to optimize experiment partitioning within a module?
        boolean gibbsSampleExpts= Boolean.parseBoolean(setProperty("gibbsSampleExpts",hp));
        // If Gibbs sampling experiments, set parameters
        int numRunsExpts=0, numStepsExpts=0, burnInExpts=0, sampleStepExpts=0, minSizeExpt=1, cutLevel=1;
        double signiFact = 1.0;
        boolean makeConsensusTree=false;
        if (gibbsSampleExpts) {
        	// number of runs
        	numRunsExpts= Integer.parseInt(setProperty("numRunsExpts",hp));
        	// number of steps in each run
        	numStepsExpts= Integer.parseInt(setProperty("numStepsExpts",hp));
        	// number of initial steps before sampling starts
        	burnInExpts= Integer.parseInt(setProperty("burnInExpts",hp));
        	// number of steps between samples
        	sampleStepExpts= Integer.parseInt(setProperty("sampleStepExpts",hp));
        	makeConsensusTree = Boolean.parseBoolean(setProperty("makeConsensusTree",hp));
        	minSizeExpt = Integer.parseInt(setProperty("minSizeExpt",hp));
        	cutLevel = Integer.parseInt(setProperty("cutLevel",hp));
        	signiFact = Double.parseDouble(setProperty("signiFact",hp));
        }

        // Score cutoff for pruning regulation trees
        double scoregain = Double.parseDouble(setProperty("scoregain",hp));

        // Use merge score from Bayesian Hierarchical Clustering or simple score difference
        // when hierarchically clustering experiments or experiment clusters?
        boolean useBHCscore = Boolean.parseBoolean(setProperty("useBHCscore",hp));

        // Assign regulators to tree nodes  stochastically or deterministically (minimum entropy)?
        boolean assignRegStoch = Boolean.parseBoolean(setProperty("assignRegStoch",hp));
        int numReg = 0;
        double betaReg = 0.0;
        boolean acyclic = false;
        if (assignRegStoch){
        	numReg = Integer.parseInt(setProperty("numReg",hp));
        	betaReg = Double.parseDouble(setProperty("betaReg",hp));
        } else {
        	acyclic = Boolean.parseBoolean(setProperty("acyclic",hp));
        }

        boolean assignAllLevels = Boolean.parseBoolean(setProperty("assignAllLevels",hp));
        int assignRegLevel = 0;
        if (!assignAllLevels){
        	assignRegLevel = Integer.parseInt(setProperty("assignRegLevel",hp));
        }
        
        // XML and score output file name, set in properties or passed as
        // second argument
        String outputFileBase;
        if (filesArgs.size() == 2){
        	outputFileBase = filesArgs.get(1);
        } else {
        	outputFileBase = setProperty("XMLfilename",hp);
        	if (outputFileBase.equals("auto")){
        		Date today = new Date();
        		DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT, Locale.GERMAN);
        		outputFileBase = "network_" + df.format(today) + "_"; 
        		df = DateFormat.getTimeInstance(DateFormat.MEDIUM, Locale.GERMAN);
        		// append random number to make sure file name is unique
        		Random rand  = new Random();
        		outputFileBase += df.format(today) + "_" + rand.nextDouble();
        	}
        }
        String xmlOutput = outputFileBase + ".xml.gz";
        
        
        String scoreFile = "";
        boolean returnScores = false;
        if (!gibbsSampleGenes) {
        	returnScores = Boolean.parseBoolean(setProperty("returnScore",hp));
        	if (returnScores){
        		scoreFile = outputFileBase  + ".score.txt";
        	}
        }
        
        if (useGUI){
//        	startupGUI.showPropertyList(100,100);
        }
        System.out.println("\nLeMoNe Configuration:");
        System.out.println("=====================");
        for (String prop : hp.keySet())      
        	System.out.println(prop + ": " +  hp.get(prop));
        System.out.println("=====================\n");

        System.out.println("XML output stored in file: " + xmlOutput + "\n");
        
        // BiNGO & graphic display of modules options
        boolean do_bingo = Boolean.parseBoolean(setProperty("BiNGO",hp));
        boolean do_draw_modules = Boolean.parseBoolean(setProperty("drawModules",hp));


        /**
         * SITUATION 1: NO INITIAL CLUSTERING, NO GIBBS SAMPLING FOR GENES AND EXPERIMENTS,
         *              ASSIGNING BEST REGULATOR TO EACH TREENODE 
         * 
         * Standard heuristic search of local optimum
         * 
         */

        if (!initClust & !gibbsSampleGenes & !gibbsSampleExpts){
        	// create module network
        	M = new ModuleNetwork(dataDir, expressiondata, numclust, 
        			clusterdata, regulatordata, lambda0, mu0, alpha0, beta0, hp);
        	// set data type
        	if (dataType.equals("AFFY"))
        		M.dataType = M.dataType.AFFY;
        	else
        		M.dataType = M.dataType.RATIO;
        	// rescale scoregain
        	scoregain = scoregain * M.numGenes;
        	// with or without storing score values?
        	if (returnScores){
        		String[] scoreFiles = new String[1];
        		scoreFiles[0] = scoreFile;
        		M.heuristicSearchMax(scoregain, useBHCscore, epsConvergence, scoreFiles);
        	} else { 
        		M.heuristicSearchMax(scoregain, useBHCscore, epsConvergence);
        	}
        	// assign regulators
        	if (acyclic){
        		M.assignRegulators();
        	} else {
        		M.assignRegulatorsNoAcycl();
        	}
        }

        /**
         * SITUATION 2: NO GENE CLUSTERING, GIBBS SAMPLING FOR EXPERIMENTS,
         *              ASSIGNING REGULATORS STOCHASTICALLY OR BY MINIMUM
         * 
         * Read an existing set of good gene clusters and generate
         * different experiment partitions each with their own regulation tree.
         * 
         */
        if (!initClust & !gibbsSampleGenes & gibbsSampleExpts){
        	// create module network
        	M = new ModuleNetwork(dataDir, expressiondata, numclust, 
        			clusterdata, regulatordata, lambda0, mu0, alpha0, beta0, hp);
        	// set data type
        	if (dataType.equals("AFFY"))
        		M.dataType = M.dataType.AFFY;
        	else
        		M.dataType = M.dataType.RATIO;
        	// rescale scoregain
        	scoregain = scoregain * M.numGenes;
        	// Gibbs sample experiments to create different hierarchical trees
        	M.gibbsSamplerExpts(numRunsExpts, numStepsExpts, burnInExpts, sampleStepExpts, 
        			scoregain, useBHCscore);
        	// make consensus trees 
        	if (makeConsensusTree)
        		M.makeConsensusTrees(cutLevel, minSizeExpt, signiFact);
        	// assign regulators
        	if (assignRegStoch)
        		// check if we need to restrict to certain level
        		if (assignAllLevels)
        			M.assignRegulatorsNoAcyclStoch(betaReg, numReg);
        		else
        			M.assignRegulatorsNoAcyclStoch(assignRegLevel, betaReg, numReg);
        	else
        		M.assignRegulatorsNoAcycl();	
        }

        /**
         * SITUATION 3: GIBBS SAMPLING FOR GENES AND EXPERIMENTS, ASSIGNING REGULATORS 
         *              STOCHASTICALLY OR BY MINIMUM
         * 
         * Find different sets of modules by iteratively Gibbs sampling genes and 
         * experiments. For each set of modules sampled, multiple hierarchical trees
         * can be sampled per module by separately Gibbs sampling experiments.
         * 
         */
        if (gibbsSampleGenes){
        	System.out.println("hey i'm here. yoooo ;)");
        	// create module network
        	M = new ModuleNetwork(dataDir, expressiondata, regulatordata, 
        			lambda0, mu0, alpha0, beta0, hp);
        	// set data type
        	if (dataType.equals("AFFY"))
        		M.dataType = M.dataType.AFFY;
        	else
        		M.dataType = M.dataType.RATIO;
        	// rescale scoregain
        	scoregain = scoregain * M.numGenes;
        	//set initial clusters if not given as an i/p to sqrt(num_genes)
        	if(initNumClust==0)
        		initNumClust=(int)Math.sqrt(M.numGenes);
        	// Gibbs sample different module sets with one tree per module
        	M.gibbsSamplerGenes(initNumClust, numRunsGenes, numStepsGenes, burnInGenes, sampleStepGenes, 
        			scoregain, useBHCscore);
        	// if Gibbs sample parameters for experiments, generate multiple trees
        	// for the different module sets
        	if (gibbsSampleExpts)
        		M.gibbsSamplerExpts(numRunsExpts, numStepsExpts, burnInExpts, sampleStepExpts, 
        				scoregain, useBHCscore);
        	// make consensus trees 
        	if (makeConsensusTree)
        		M.makeConsensusTrees(cutLevel, minSizeExpt, signiFact);
        	// assign regulators to all trees for all modules for all moduleSets
        	if (assignRegStoch)
        		// check if we need to restrict to certain level
        		if (assignAllLevels)
        			M.assignRegulatorsNoAcyclStoch(betaReg, numReg);
        		else
        			M.assignRegulatorsNoAcyclStoch(assignRegLevel, betaReg, numReg);
        	else
        		M.assignRegulatorsNoAcycl();
        }

        
        //write XML file (gzipped)
        M.writeXML(xmlOutput);

        // write txt output with module sets
        M.printInModules(outputFileBase + "_moduleSets.txt");


        /*  run BiNGO (GO statistics) on all modules and for each module set
         *  parameters are taken from the property file, for example:
         *  
         *  #BiNGO parameters
         *  BiNGO=true
         *  signif_def=0.05
         *  annotation_file_def=A_thaliana_default
         *  ontology_file_def=GO_Biological_Process
         *  annotation_default=true
         *  ontology_default=true
         * 
         */

        if (do_bingo) {
//            String signif_def = setProperty("signif_def",hp);
//            String annotation_file_def = setProperty("annotation_file_def",hp);
//            String ontology_file_def = setProperty("ontology_file_def",hp);
//            boolean annotation_default = Boolean.parseBoolean(setProperty("annotation_default",hp));
//            boolean ontology_default = Boolean.parseBoolean(setProperty("ontology_default",hp));
//        	//BiNGO b = new BiNGO(M);
//            System.out.println("BiNGO functional annotation...");
//            BiNGO b = new BiNGO(M,"notused","notused", annotation_file_def, ontology_file_def, signif_def, annotation_default, ontology_default);
        }

        // draw modules genes + regulation program
        if (do_draw_modules) {
        	for (Module mod : M.moduleSet){
        		String filename = "module" + mod.number + ".png";
        		mod.draw(filename);
        	}	
        }
  
    }
    
   
    /**
     * 
     * @param filename name of properties file, see <code>LeMoNe.properties</code> for an example.
     * @param h HashMap of (key, value) pairs for each property.
     * @author erbon
     */
    public static void getProperties(String filename, HashMap<String,String> h) {
        try {
            
            BufferedReader file = new BufferedReader(new FileReader(filename));
            String inputline;
            
            // Read data from file
            while ((inputline = file.readLine()) != null) {
                inputline = inputline.trim();
                if (!inputline.startsWith("#")) {
                    Scanner s = new Scanner(inputline).useDelimiter("=");
                    if (s.hasNext()) {
                        String key = s.next().trim();
                        String value = s.next().trim();
                        //System.out.println(key + "<=>" + value);
                        h.put(key,value);
                    }
                }
            }
            file.close();
        } catch (IOException e){
        	
        }
    }
    
    /**
     * Extract property key with given name from HashMap, abort program if key is not found.
     * @param name name of key
     * @param h HashMap of properties
     * @return value of key
     * @author erbon
     */
    public static String setProperty (String name,HashMap<String,String> h) {
        String ret;
        if (h.get(name) == null)
//        	throw new Exception("property " + name + " not found in parameter file.");
            Die("property " + name + " not found in parameter file.");
        
        ret = h.get(name);
        return(ret);
    }
    
    /**
     * Abort program with error message.
     * @param str error message
     * @author erbon
     */
    public static void Die(String str) {
        System.out.println("Error: " + str);
        System.exit(1);
    }
    
}
