function [p,s,m] = fuzzyclust(s)
% FUZZYCLUST - Fuzzy clustering of pairwise cluster probabilities
% fuzzyclust converts a matrix whose entries are the probability
% for a pair of genes to be clustered together to a matrix p of 
% soft clusters, such that p(i,k) is the probability that i belongs 
% to cluster k. The number of fuzzy clusters is determined
% automatically. 
%
% USAGE:
%  [p,s,m] = fuzzyclust(s)
%    Input:  s NxN pairwise probability matrix
%    Output: p NxK soft cluster probabilities, s truncated input matrix,
%            m = sum(p,2) to measure what fraction of genes has been
%            assigned
%
% Written by
%   Anagha Joshi - anagha.joshi@psb.ugent.be
%   Tom Michoel - tom.michoel@psb.ugent.be
%
% For more information, see our paper:
%   A. Joshi, Y. Van de Peer, T. Michoel; Analysis of a Gibbs sampler
%      method for model based clustering of gene expression data,
%      Bioinformatics (accepted) (2007). 

opteigs.disp = 0;
p = sparse(length(s),1);
m = p;
nondeg = true;
dim = length(s);

% loop untill s consists of numerical noise only or has a degenerate
% largest eigenvalue
while (max(max(s))>eps & nondeg)
  k = size(p,2);
%  disp([k  full(max(max(s)))])
  [i,j,s]=find(s);
  a=sqrt(1-m);
  s = a(i).*a(j).*s;
  s = sparse(i,j,s,dim,dim);
  [v,d] = eigs(s,2,'lm',opteigs);
  v = sparse(v(:,1));
  d = diag(d);
  % check degeneracy
  if (abs(d(1)-d(2))<eps)
%    disp('degenerate eigenvalue');
%    disp([d(1) d(2)])
    nondeg = false;
  else
    % create v with positive entries, upto  numerical precision
    [vmax, imax] = max(abs(v));
    v = v./v(imax); % this one should have all positive elements
    v(find(v<0))=0;
    p = [p min((1-m(imax)).*v,1-m)];
    m=m+p(:,end);
  end
end

% first column of p is empty
p = p(:,2:end);